import 'dart:async';

import 'package:flutter/widgets.dart';
import 'package:provider/provider.dart';
import 'package:sunny_dart/sunny_dart.dart' hide Consumer;
import 'package:sunny_forms/stateful_route.dart';

typedef EventInvoker<T> = FutureOr<T> Function();

/// A callback callback.  Allows an external source to pass in hook, so that the internal code can provide what execution
/// is supposed to occur when that hook is fired.
typedef EventInvokerRegistrar<T> = void Function(
    dynamic self, EventInvoker<T> invoker);
typedef ContextEventInvokerRegistrar<T> = EventInvokerRegistrar<T> Function(
    BuildContext context);

typedef CommandBuilder<T> = Widget Function(
    BuildContext context, CallBridge<T> command);

/// This widget is used to connect two sibling components, and allow one to trigger an action in the other.
class Command<T> extends StatelessWidget with SunnyStatelessRoutingMixin<T> {
  final CommandBuilder<T> builder;
  final String name;

  Command(this.name, this.builder) : super(key: Key("bridger-$name"));

  @override
  Widget build(BuildContext context) {
    return Provider(
      key: Key("provider-bridger-$name"),
      create: (context) => CallBridge<T>(name),
      child: Consumer<CallBridge<T>>(
        builder: (BuildContext context, CallBridge<T> value, Widget? child) {
          return builder(context, value);
        },
      ),
    );
  }
}

class CallBridge<T> with LoggingMixin {
  final String debugName;
  EventInvoker<T>? _call;
  dynamic _registrar;

  CallBridge(this.debugName);

  void onEvent(self, EventInvoker<T> invoker) {
    if (_call != null && _registrar != self) {
      throw "Already have a call from $_registrar (new registrar is $self)";
    }
    this._call = invoker;
    this._registrar = self;
  }

  void unregisterCall(self) {
    if (this._registrar == self) {
      this._call = null;
      this._registrar = null;
    } else {
      throw "Wrong invoker expected $_registrar";
    }
  }

  FutureOr<T> call() {
    if (_call == null) {
      log.severe(
          "No call registered for $debugName $T", null, StackTrace.current);
      throw "No call registered for $debugName $T";
    }
    return _call!();
  }
}

extension BuildContextBridgeExtensions on BuildContext {
  FutureOr<T> invokeBridge<T>() {
    return Provider.of<CallBridge<T>>(this, listen: false).call();
  }

  void onBridge<T>(self, EventInvoker<T> invoker) {
    return Provider.of<CallBridge<T>>(this, listen: false)
        .onEvent(self, invoker);
  }

  CallBridge<T>? bridge<T>() {
    try {
      return Provider.of<CallBridge<T>>(this, listen: false);
    } on ProviderNotFoundException {
      return null;
    }
  }

  void unregisterOnBridge<T>(self) {
    return Provider.of<CallBridge<T>>(this, listen: false).unregisterCall(self);
  }

  EventInvokerRegistrar<T> caller<T>() {
    return Provider.of<CallBridge<T>>(this, listen: false).onEvent;
  }
}
