import 'package:equatable/equatable.dart';
import 'package:flexidate/flexidate.dart';
import 'package:flutter/cupertino.dart';
import 'package:flutter/material.dart';
import 'package:flutter/services.dart';
import 'package:flutter/widgets.dart';
import 'package:flutter_typeahead/flutter_typeahead.dart';
import 'package:sunny_dart/helpers.dart';
import 'package:sunny_dart/helpers/logging_mixin.dart';
import 'package:sunny_dart/sunny_get.dart';
import 'package:sunny_dart/typedefs.dart';
import 'package:sunny_forms/clear_input.dart';
import 'package:sunny_forms/form_fields/select_field.dart';
import 'package:sunny_forms/form_fields/typeaheads.dart';
import 'package:sunny_forms/raw_text_field.dart';
import 'package:sunny_forms/smart_dates/smart_date_query.dart';
import 'package:sunny_forms/smart_dates/smart_date_service.dart';
import 'package:sunny_forms/sunny_form_controller.dart';
import 'package:sunny_forms/sunny_form_field.dart';
import 'package:sunny_forms/text_widgets.dart';
import 'package:sunny_forms/widget_extensions.dart';
import 'package:sunny_sdk_core/model.dart';
import 'package:sunny_sdk_core/services.dart';
import 'package:sunny_sdk_core/services/sunny.dart';
import 'package:sunny_service_stubs/meta.dart';
import 'package:sunny_service_stubs/services.dart';
import 'package:sunny_service_stubs/sunny_service_stubs.dart';

class DateTypeaheadControl extends StatefulWidget {
  final FormValueChanged<SmartDateQueryResult?>? onChange;
  final dynamic initialValue;

  /// If the text of the selection is different than the formatted date
  final String? initialSelection;
  final String? placeholder;
  final String? label;
  final bool? autofocus;
  final TextInputAction? textInputAction;
  final TextEditingController? textController;
  final bool? enabled;
  final FocusNode? focusNode;

  final SmartDateOptionsConfig inputContext;

  const DateTypeaheadControl({
    Key? key,
    this.onChange,
    this.initialValue,
    this.placeholder,
    this.autofocus = false,
    this.label,
    this.enabled = true,
    this.textInputAction,
    this.focusNode,
    this.textController,
    SmartDateOptionsConfig? inputContext,
    this.initialSelection,
  })  : inputContext = inputContext ?? const SmartDateOptionsConfig.flexiDate(),
        super(key: key);

  /// Convenience for the different form field constructors
  static DateTypeaheadControl ofFormField<T extends Object>({
    String? placeholder,
    required SunnyFormFieldState<T> state,
    required Mapping<T?, SmartDateQueryResult?> toResult,
    required Mapping<SmartDateQueryResult?, T?> fromResult,
    required Formatter<T?> formatter,
    bool? autofocus,
    SmartDateOptionsConfig? inputContext,
    TextInputAction? textInputAction,
  }) {
    return DateTypeaheadControl(
        onChange: ((value, from) => state.updateValue(fromResult(value), from)),
        initialValue: toResult(state.value),
        initialSelection: formatter(state.value),
        autofocus: autofocus,
        focusNode: state.focusNode,
        inputContext: inputContext,
        key: Key("dateTypeahead-${state.attribute.toKey()}"));
  }

  factory DateTypeaheadControl.ofDateTimeField({
    required SunnyFormFieldState<DateTime> state,
    String? placeholder,
    bool autofocus = false,
    TextInputAction? textInputAction,
    SmartDateOptionsConfig? inputContext,
  }) =>
      DateTypeaheadControl.ofFormField<DateTime>(
          state: state,
          autofocus: autofocus,
          fromResult: (result) => result?.resultDate,
          placeholder: placeholder ?? "e.g. April 6, 2013",
          textInputAction: textInputAction,
          toResult: (dateTime) => dateTime == null ? null : SmartDateQueryResult.ofDateTime(dateTime),
          inputContext: inputContext ?? const SmartDateOptionsConfig.flexiDate(),
          formatter: (dateTime) => smartDateFormat(dateTime) ?? '');

  factory DateTypeaheadControl.ofFlexiDate({
    required SunnyFormFieldState<FlexiDate> state,
    String? placeholder,
    bool autofocus = false,
    TextInputAction? textInputAction,
    SmartDateOptionsConfig? inputContext,
  }) =>
      DateTypeaheadControl.ofFormField<FlexiDate>(
        state: state,
        autofocus: autofocus,
        fromResult: (result) => result?.flexiDate,
        placeholder: placeholder ?? "e.g. April 6, 2013",
        textInputAction: textInputAction,
        toResult: (dateTime) => dateTime == null ? null : SmartDateQueryResult.ofFlexiDate(dateTime),
        inputContext: inputContext ?? const SmartDateOptionsConfig.flexiDate(),
        formatter: (flexi) => flexi?.formatted() ?? '',
      );

  @override
  _DateTypeaheadControlState createState() => _DateTypeaheadControlState();
}

class _DateTypeaheadControlState extends State<DateTypeaheadControl> with LoggingMixin {
  late TextEditingController? _textController;

  String? id;

  @override
  void initState() {
    super.initState();
    id = uuid();
    String? startValue = widget.initialSelection;
    final initialValue = widget.initialValue;
    if (initialValue is FlexiDate) {
      startValue ??= initialValue.formatted();
    } else if (initialValue is DateTime) {
      startValue ??= smartDateFormat(tzDateTime(initialValue));
    }
    startValue ??= "";
    _textController = widget.textController ??
        TextEditingController.fromValue(
          TextEditingValue(
            text: startValue,
            selection: TextSelection.collapsed(offset: startValue.length),
          ),
        );
  }

  @override
  dispose() {
    if (widget.textController == null) {
      _textController!.dispose();
    }

    super.dispose();
  }

  @override
  Widget build(BuildContext context) {
    final textFieldConfig = getTypeaheadInputConfig(
      context,
      placeholder: widget.placeholder,
      inputAction: widget.textInputAction,
      focusNode: widget.focusNode,
      controller: _textController,
      autofocus: widget.autofocus,
      hasPrefix: true,
      hasSuffix: true,
    );

    return inputWithIcon(
      prefix: context.icons.calendar.icon,
      suffix: ClearInputIcon(
        key: Key("clearInput-$id"),
        focusNode: widget.focusNode,
        notifier: widget.onChange,
        textEditingController: _textController,
        showExpand: false,
        icon: Icon(Icons.clear),
      ),
      color: widget.focusNode?.hasFocus == true || widget.initialValue != null ? sunnyColors.primaryColor : sunnyColors.textLight,
      input: TypeAheadField<SmartDateQueryResult>(
        autoFlipDirection: true,
        textFieldConfiguration: textFieldConfig,
        keepSuggestionsOnSuggestionSelected: false,
        noItemsFoundBuilder: (context) {
          return buildNoItemsTile(context, message: "No results");
        },
        suggestionsCallback: (query) async {
          final location = sunnyLocalization.userLocation;
          final smartDateService = sunny.get<SmartDateService>();
          final results = await smartDateService.parseDates(
            query,
            isFlexiDate: widget.inputContext.isFlexiDate,
            contact: widget.inputContext.contact,
            facts: widget.inputContext.facts,
            location: location,
          );
          return results;
        },
        errorBuilder: typeaheadErrorBuilder,
        animationStart: 1.0,
        transitionBuilder: (context, widget, controller) {
          return widget;
        },
        itemBuilder: (BuildContext context, SmartDateQueryResult itemData) {
          return TypeaheadOption(child: itemData.buildListTile(context));
        },
        hideOnLoading: false,
        hideOnEmpty: false,
        hideOnError: false,
        hideSuggestionsOnKeyboardHide: false,
        onSuggestionSelected: (SmartDateQueryResult suggestion) async {
          _textController!.text = suggestion.selection!;

          final result = await suggestion.onSelected(context);
          if (result != null) {
            widget.onChange!(result, AttributeSource.control);
            _textController!.text = result.selection!;
          }
          return;
        },
      ),
    );
  }
}

/// The parameters passed as input to the date input
class SmartDateOptionsConfig extends Equatable {
  final bool _isFlexiDate;
  final IContact? contact;
  final List<IFact>? _facts;
  final List<DateRef>? _defaultDates;
  final DateRef? defaultDate;
  final bool? _showAlert;

  const SmartDateOptionsConfig.of({
    required bool isFlexiDate,
    required this.contact,
    required List<IFact>? facts,
    required List<DateRef>? defaultDates,
    required this.defaultDate,
    required bool? showAlert,
  })  : _showAlert = showAlert,
        _isFlexiDate = isFlexiDate,
        _facts = facts,
        _defaultDates = defaultDates;

  const SmartDateOptionsConfig.flexiDate({
    IContact? contact,
    List<IFact> facts = const [],
    List<DateRef>? defaultDates,
    DateRef? defaultDate,
    bool? showAlert,
  }) : this.of(
          isFlexiDate: true,
          contact: contact,
          facts: facts,
          defaultDates: defaultDates,
          defaultDate: defaultDate,
          showAlert: showAlert,
        );

  @override
  List<Object?> get props => [
        _isFlexiDate,
        contact,
        _facts,
        _defaultDates,
        defaultDate,
        _showAlert,
      ];

  List<DateRef> get defaultDates => _defaultDates ?? const [];

  bool get isFlexiDate => _isFlexiDate;

  bool get showAlert => _showAlert ?? false;

  List<IFact> get facts => _facts ?? const [];

  SmartDateOptionsConfig operator +(SmartDateOptionsConfig other) {
    return this.copy(
      isFlexiDate: other._isFlexiDate,
      contact: other.contact,
      facts: other._facts,
      defaultDates: other._defaultDates,
      defaultDate: other.defaultDate,
      showAlert: other._showAlert,
    );
  }

  SmartDateOptionsConfig copy({
    bool? isFlexiDate,
    IContact? contact,
    List<IFact>? facts,
    List<DateRef>? defaultDates,
    DateRef? defaultDate,
    bool? showAlert,
  }) {
    return SmartDateOptionsConfig.of(
      isFlexiDate: isFlexiDate ?? this._isFlexiDate,
      contact: contact ?? this.contact,
      facts: facts ?? this._facts,
      defaultDates: defaultDates ?? this._defaultDates,
      defaultDate: defaultDate ?? this.defaultDate,
      showAlert: showAlert ?? this._showAlert,
    );
  }
}
