import 'package:flutter/cupertino.dart';
import 'package:flutter/foundation.dart';
import 'package:flutter/material.dart';
import 'package:flutter/widgets.dart';
import 'package:pfile/pfile.dart';
import 'package:sunny_core_widgets/sunny_core_widgets.dart';
import 'package:sunny_dart/helpers/safe_completer.dart';
import 'package:sunny_dart/sunny_dart.dart';
import 'package:sunny_dart/typedefs.dart';
import 'package:sunny_essentials/state.dart';
import 'package:sunny_forms/form_fields/media_service.dart';
import 'package:sunny_forms/icon/image_widgets.dart';
import 'package:sunny_service_stubs/models.dart';
import 'package:sunny_service_stubs/services.dart';

import '../media.dart';

/// A control for uploading image urls
class ImageUrlControl extends StatefulWidget {
  final Uri? initialValue;
  final Widget? placeholder;
  final NamedIcon? overlay;
  final VoidCallback? onStartChange;
  final VoidCallback? onEndChange;
  final ValueChanged<Uri>? onUploaded;
  final ValueChanged<PFile?>? onFileChanged;
  final ErrorCallback? onError;
  final double size;
  final String? type;
  final Producer<String>? imageId;
  final bool isReadOnly;
  final bool allowThirdParty;
  final bool hideOverlay;
  final bool preemptPermission;

  /// Used if any callers want to be notified when this component is ready to be  processed
  final SafeCompleter? readyCheck;

  const ImageUrlControl({
    Key? key,
    this.initialValue,
    this.onUploaded,
    this.onError,
    this.readyCheck,
    this.size = 60,
    this.onStartChange,
    this.onEndChange,
    this.type,
    this.onFileChanged,
    this.hideOverlay = false,
    this.imageId,
    this.placeholder,
    this.isReadOnly = false,
    this.preemptPermission = false,
    this.overlay,
    this.allowThirdParty = false,
  }) : super(key: key);

  @override
  State<StatefulWidget> createState() => ImageUrlControlState();
}

class ImageUrlControlState extends BaseState<ImageUrlControl> {
  Uri? _imageUri;
  ChosenMedia? _imageFile;

  set imageFile(ChosenMedia file) {
    safeState(() {
      _imageFile = file;
    });
  }

  set imageUri(Uri uri) {
    safeState(() {
      _imageUri = uri;
    });
  }

  @override
  void initState() {
    super.initState();
    _imageUri = widget.initialValue;
    processing?.complete(true);
  }

  void dispose() {
    processing?.complete(false);

    super.dispose();
  }

  SafeCompleter? get processing => widget.readyCheck;

  @override
  Widget build(BuildContext context) {
    return Tappable(
      onTap: widget.isReadOnly
          ? null
          : (context) async {
              widget.onStartChange?.call();
              try {
                final newImageFiles = await showMediaListPicker(
                    context, MediaContentType.image,
                    preemptPermission: widget.preemptPermission,
                    crop: !kIsWeb,
                    onLoading: () {},
                    onChange: (file, bool done) {});

                if (newImageFiles.isNullOrEmpty) return null;

                imageFile = newImageFiles!.first;
                if (_imageFile!.file != null) {
                  widget.onFileChanged?.call(newImageFiles.first.file);
                  final String imageId =
                  (await (widget.imageId?.call()) ?? uuid());
                  try {
                    processing?.reset();
                    processing?.start();

                    final tracker = await mediaService.uploadImage(
                        _imageFile!.file,
                        pictureType: widget.type,
                        pictureId: imageId);

                    final url = await tracker.result;
                    log.info("Uploaded image: $url");
                    imageUri = url;
                    if (processing != null) {
                      log.info("Marking waitRequest as complete");
                      processing?.complete();
                    }
                    widget.onUploaded?.call(url);
                  } catch (e, stack) {
                    if (processing != null) {
                      log.info(
                          "Marking waitRequest as complete due to error $e");
                      processing?.complete();
                    }
                    log.severe("Error $e", e, stack);
                    widget.onError?.call(e, stack);
                  }
                } else {
                  final uri =
                  (_imageFile!.content as IImageContent).imageUrl?.toUri();
                  if (uri != null) {
                    processing?.complete();
                    imageUri = uri;
                    widget.onUploaded?.call(uri);
                  }
                }
              } finally {
                widget.onEndChange?.call();
              }
            },
      child: SunnyPad(
        child: imageWidgetOf(
          context,
          [_imageFile, _imageUri],
          key: Key("image-picker"),
          size: widget.size,
          circle: true,
          iconOverlay: widget.isReadOnly
              ? widget.overlay
              : (widget.hideOverlay
                  ? null
                  : ClipOval(
                      child: Container(
                      color: sunnyColors.textLight,
                      height: 22,
                      width: 22,
                      child: Icon(
                        NamedIcons.camera,
                        size: 13,
                        color: Colors.black,
                      ),
                    ))),
          //grayCircular(),
          placeholder: _imageFile != null
              ? null
              : widget.placeholder ??
                  ClipOval(
                    child: Container(
                      alignment: AlignmentDirectional.center,
                      height: widget.size,
                      width: widget.size,
                      color: sunnyColors.primaryColor,
                      child: NamedIcons.upload.build(
                          color: Colors.white, size: (widget.size * .75)),
                    ),
                  ),
        ),
      ),
    );
  }
}
