import 'dart:async';

import 'package:collection_diff/collection_diff.dart';
import 'package:flutter/widgets.dart';
import 'package:flutter_chips_input_sunny/flutter_chips_input.dart';
import 'package:sunny_dart/sunny_dart.dart';
import 'package:sunny_essentials/state.dart';
import 'package:sunny_forms/options/options_service.dart';
import 'package:sunny_forms/platform/platform_chip.dart';
import 'package:sunny_forms/sunny_form_controller.dart';
import 'package:sunny_forms/sunny_form_field.dart';
import 'package:sunny_forms/widget_extensions.dart';
import 'package:sunny_service_stubs/options.dart';

/// Builds a list of items using a chip input
class ListSelectControl<K, T> extends StatefulWidget {
  final String? selectedType;
  final String id;
  final FormValueChanged<List<KeyedOption<K, T>?>> onChange;
  final Iterable<K>? initialValue;
  final KeyedOptionsHandler<K, T> handler;
  final TypeaheadHandler<K, T> viewHandler;
  final String? placeholder;
  final FocusNode? focusNode;
  final bool isRequired;

  ListSelectControl.ofValue({
    Key? key,
    required this.handler,
    required this.viewHandler,
    required this.id,
    this.selectedType,
    FormValueChanged<List<T?>>? onChange,
    this.isRequired = false,
    Iterable<T>? initialValue,
    this.placeholder,
    this.focusNode,
  })  : onChange = ((options, source) {
          // Unbox the option as values
          final mapped = options.map((o) => o?.value).notNullList().toList();
          onChange?.call(mapped, source);
        }),
        initialValue = initialValue?.map(handler.toKey).notNull(),
        super(key: key);

  ListSelectControl.ofKey({
    Key? key,
    required this.handler,
    required this.viewHandler,
    required this.id,
    this.selectedType,
    FormValueChanged<List<K?>>? onChange,
    this.isRequired = false,
    Iterable<K>? initialValue,
    this.placeholder,
    this.focusNode,
  })  : onChange = ((options, source) {
          final mapped = options.map((o) => o?.key).notNull().toList();
          onChange?.call(mapped, source);
        }),
        initialValue = [...?initialValue],
        super(key: key);

  @override
  _ListSelectControlState<K, T> createState() =>
      _ListSelectControlState<K, T>();
}

class _ListSelectControlState<K, T> extends BaseState<ListSelectControl<K, T>> {
  ChipsInputController<KeyedOption<K, T>?>? listController;

  late KeyedOptionsHandler<K, T> _handler;

  @override
  void initState() {
    super.initState();
    _handler = widget.handler;
    listController = ChipsInputController(
      findSuggestions: _findItems,
      equality: DiffEquality.ofEquality(),
      tokenizer: (item) => [],
    );
    _init();
  }

  _init() async {
    final mapped = await (handler
        .resolveAllOptions(context, widget.initialValue)
        .awaitAll() as FutureOr<List<KeyedOption<K, T>>>);
    await listController!
        .syncChips(mapped, source: ChipChangeOperation.external);
  }

  void dispose() {
    listController!.dispose();
    super.dispose();
  }

  KeyedOptionsHandler<K, T> get handler => _handler;

  @override
  Widget build(BuildContext context) {
    return ChipsInput<KeyedOption<K, T>?>(
      id: widget.id,
      placeholder: widget.placeholder,
      chipId: (k) => "${k!.key}",
      enabled: true,
      focusNode: widget.focusNode,
      controller: listController,
      suggestionBuilder: (context, controller, index, KeyedOption<K, T>? chip) {
        _onSelected(KeyedOption<K, T>? chip) async {
          if (chip is KeyedAdhocOption<K, T> && chip.hasAdhocCreator) {
            final result = await chip.runAdhocCreator(context);
            if (result != null) {
              Future.delayed(200.ms, () async {
                if (mounted) {
                  final newOption = handler.toOption(result);
                  _onSelected(newOption);
                }
              });
            }
          } else if (chip is KeyedContextSwitchOption<K, T>) {
            _handler = chip.onSelect() as KeyedOptionsHandler<K, T>;
            listController!.loadSuggestions(listController!.query);
          } else {
            if (mounted) {
              setState(() {
                controller!.addChip(chip, resetQuery: true);
              });
            }
          }
        }

        return widget.viewHandler
            .wrapSuggestionTile(widget.viewHandler.renderSuggestionTile(
          context,
          chip!,
          selectOption: _onSelected,
          onTap: () async {
            return _onSelected(chip);
          },
          isSelected: false,
        ));
      },
      onChipsChanged: (control, source) {
        AttributeSource attrSource;
        switch (source) {
          case ChipChangeOperation.external:
            attrSource = AttributeSource.model;
            break;
          case ChipChangeOperation.addChip:
            attrSource = AttributeSource.control;
            break;
          case ChipChangeOperation.deleteChip:
            attrSource = AttributeSource.control;
            break;
          case ChipChangeOperation.enterDirect:
            attrSource = AttributeSource.control;
            break;
          default:
            attrSource = AttributeSource.model;
            break;
        }
        widget.onChange.call([...?control?.chips], attrSource);
      },
      chipBuilder: (BuildContext context, controller, int index,
          KeyedOption<K, T>? chip) {
        return PlatformChip(
            key: Key("${chip!.diffKey}"),
            onDeleted: () {
              controller!.removeAt(index, resetQuery: true);
            },
            label: Text(chip.selection));
      },
    );
  }

  Future<ChipSuggestions<KeyedOption<K, T>?>> _findItems(String query) async {
    try {
      final filtered = await _handler.filter(query, listController!.chips, 10);
      if (query.isNotNullOrBlank &&
          handler.adhocOptionBuilder != null &&
          filtered.isEmpty) {
        final adhocOption = handler.adhocOptionBuilder?.call(query);
        if (adhocOption == null) return const ChipSuggestions.empty();
        return ChipSuggestions<KeyedOption<K, T>>(suggestions: [adhocOption]);
      }
      return ChipSuggestions<KeyedOption<K, T>?>(
          suggestions: filtered.map((i) => handler.toOption(i)).toList());
    } catch (e) {
      log.warning("Error generating suggestions: $e");
      return ChipSuggestions<KeyedOption<K, T>>(suggestions: []);
    }
  }
}
