import 'package:flutter/cupertino.dart';
import 'package:flutter/material.dart';
import 'package:flutter/services.dart';
import 'package:flutter/widgets.dart';
import 'package:flutter_typeahead/flutter_typeahead.dart';

import 'package:dartxx/dartxx.dart';
import 'package:sunny_dart/helpers/logging_mixin.dart';
import 'package:sunny_dart/typedefs.dart';
import 'package:sunny_forms/clear_input.dart';
import 'package:sunny_forms/form_fields/typeaheads.dart';
import 'package:sunny_forms/raw_text_field.dart';
import 'package:sunny_forms/sunny_form_controller.dart';
import 'package:sunny_forms/sunny_form_field.dart';
import 'package:sunny_forms/sunny_forms.dart';
import 'package:sunny_forms/widget_extensions.dart';
import 'package:sunny_service_stubs/models.dart';
import 'package:sunny_service_stubs/services.dart';

/// A duration typeahead - control-only
class PhysicalLocationTypeaheadControl extends StatefulWidget {
  final FormValueChanged<IPhysicalLocation?>? onChange;
  final IPhysicalLocation? initialValue;
  final String? placeholder;
  final String? label;
  final bool autofocus;
  final TextInputAction? textInputAction;
  final TextEditingController? textController;
  final bool enabled;
  final FocusNode? focusNode;
  final Formatter<IPhysicalLocation>? formatter;
  final bool? showIcon;

  const PhysicalLocationTypeaheadControl({
    Key? key,
    this.onChange,
    this.showIcon,
    this.initialValue,
    this.placeholder,
    this.autofocus = false,
    this.label,
    this.enabled = true,
    this.textInputAction,
    this.focusNode,
    this.textController,
    this.formatter,
  }) : super(key: key);

  @override
  _PhysicalLocationTypeaheadControlState createState() =>
      _PhysicalLocationTypeaheadControlState();
}

class _PhysicalLocationTypeaheadControlState
    extends State<PhysicalLocationTypeaheadControl> with LoggingMixin {
  TextEditingController? _textController;

  @override
  void initState() {
    super.initState();

    final startValue = widget.initialValue?.toDescription() ?? "";
    _textController = widget.textController ??
        TextEditingController.fromValue(
          TextEditingValue(
            text: startValue,
            selection: TextSelection.collapsed(offset: startValue.length),
          ),
        );
  }

  @override
  dispose() {
    if (widget.textController == null) {
      _textController!.dispose();
    }
    super.dispose();
  }

  @override
  Widget build(BuildContext context) {
    final formatter =
        widget.formatter ?? ((location) => location.toDescription() ?? "");
    final textFieldConfig = getTypeaheadInputConfig(
      context,
      placeholder: widget.placeholder,
      inputAction: widget.textInputAction,
      focusNode: widget.focusNode,
      controller: _textController,
      autofocus: widget.autofocus,
      hasPrefix: widget.showIcon != false,
      hasSuffix: true,
    );

    return inputWithIcon(
      prefix: widget.showIcon != false ? NamedIcons.location.icon : null,
      suffix: ClearInputIcon(
        icon: context.icons.clear.icon,
        notifier: widget.onChange,
        focusNode: widget.focusNode,
        textEditingController: _textController,
        showExpand: false,
      ),
      color: widget.focusNode?.hasFocus == true || widget.initialValue != null
          ? sunnyColors.primaryColor
          : sunnyColors.textLight,
      input: TypeAheadField<IPhysicalLocation>(
        autoFlipDirection: true,
        textFieldConfiguration: textFieldConfig,
        loadingBuilder: (context) {
          return SizedBox(height: 30, width: 30, child: loadingIndicator);
        },
        noItemsFoundBuilder: (context) {
          return buildNoItemsTile(context, message: "Enter a location");
        },
        suggestionsCallback: (query) {
          return searchPlaceInGoogle(query);
        },
        errorBuilder: typeaheadErrorBuilder,
        animationStart: 0.25,
        transitionBuilder: (context, widget, controller) {
          return widget;
        },
        itemBuilder: (BuildContext context, IPhysicalLocation itemData) {
          return PlatformListTile(
            leading: NamedIcons.location.icon.reIcon(size: 20),
            padding: EdgeInsets.all(5),
            margin: EdgeInsets.only(bottom: 8),
            useShadow: false,
            title: itemData.description == null
                ? null
                : Text(itemData.description!),
            subtitle: itemData.type == null ? null : Text(itemData.type!),
          );
        },
        keepSuggestionsOnLoading: true,
        hideOnEmpty: false,
        hideOnError: false,
        hideSuggestionsOnKeyboardHide: false,
        onSuggestionSelected: (IPhysicalLocation suggestion) {
          if (suggestion.countryCode == null) {
            // is it fully loaded?
            _textController!.text =
                "${formatter(suggestion)} (loading details)";
            lookupPlaceInGoogle(suggestion.googlePlaceId).then((p) {
              var res = IPhysicalLocation.of()
                ..fromGooglePlaceDetails(p!, placeId: suggestion.googlePlaceId);
              widget.onChange!(res, AttributeSource.control);
              _textController!.text = "${formatter(suggestion)}";
            });
          } else {
            // is it fully loaded?
            _textController!.text = formatter(suggestion);
            widget.onChange!(suggestion, AttributeSource.control);
          }
          return;
        },
      ),
    );
  }
}

extension _PhysicalLocationExts on IPhysicalLocation {
  String? toDescription() {
    final description = this.description;
    if (description.isNullOrBlank &&
        description != locality &&
        description != region) {
      if (countryCode.isNotNullOrBlank && countryCode != "US") {
        return [description, countryCode].join(", ");
      } else {
        return description;
      }
    } else if (countryCode.isNotNullOrBlank && countryCode != "US") {
      return [locality, countryCode].join(", ");
    } else {
      return [locality, region].join(", ");
    }
  }
}
