import 'dart:async';
import 'dart:math' as math;

import 'package:dartxx/dartxx.dart';
import 'package:flutter/cupertino.dart';
import 'package:flutter/foundation.dart';
import 'package:flutter/material.dart';
import 'package:flutter/widgets.dart';
import 'package:reorderables2/reorderables2.dart';
import 'package:sunny_dart/sunny_dart.dart'
    hide SunnyIterableExtensions, SunnyIterableExtensionExt;
import 'package:sunny_essentials/state.dart';
import 'package:sunny_forms/form_fields/media_service.dart';
import 'package:sunny_forms/media.dart';
import 'package:sunny_forms/media/media_content_type.dart';
import 'package:sunny_forms/media/media_picker.dart';
import 'package:sunny_forms/sunny_hud.dart';
import 'package:sunny_forms/text_widgets.dart';

import 'media_source.dart';

/// A control for uploading image urls
class MediaGalleryControl<F> extends StatefulWidget {
  final double? itemHeight;
  final int columns;
  final String name;
  final List<F>? initialValue;
  final MediaContentType<F> type;
  final Widget? placeholder;
  final ValueChanged<List<F>>? onChanged;
  final ErrorCallback? onError;
  final double size;
  final SafeCompleter? progress;
  final bool isReadOnly;
  final bool allowThirdParty;

  const MediaGalleryControl({
    Key? key,
    required this.type,
    required this.name,
    this.initialValue,
    this.onChanged,
    this.onError,
    this.columns = 3,
    this.itemHeight,
    this.allowThirdParty = false,
    this.size = 60,
    this.placeholder,
    this.isReadOnly = false,
    this.progress,
  }) : super(key: key);

  @override
  State<StatefulWidget> createState() => MediaGalleryControlState<F>();
}

class MediaGalleryControlState<F> extends BaseState<MediaGalleryControl<F>> {
  List<MediaSource<F>>? _mediaUris;

  SafeCompleter? processing;
  ScrollController? _wrap;

  StreamController<String>? completions;

  set mediaUris(Iterable<MediaSource<F?>> uris) {
    final trimmed = uris.whereType<MediaSourceData<F>>();
    var newValues = [
      ...trimmed,
      MediaSourcePlaceholder<F>(),
    ];
    if (mounted) {
      setState(() {
        _mediaUris = newValues;
      });
      notifyCompleted();
    }
  }

  void notifyCompleted() {
    var completedItems = [
      for (var u in _mediaUris.orEmptyIter().whereType<MediaSourceData<F>>())
        if (u.isCompleted && u.mediaContent != null) u.mediaContent,
    ];
    widget.onChanged?.call(completedItems.notNullList());
  }

  @override
  void initState() {
    super.initState();
    _wrap = ScrollController();
    processing = widget.progress ?? SafeCompleter.stopped();
    if (processing!.isStarted) {
      processing!.complete();
    }
    _mediaUris = [
      ...?widget.initialValue?.map((content) =>
          MediaSourceData<F>.ofMediaContent(widget.type, content,
              mediaId: null)),
      const MediaSourcePlaceholder(),
    ];
  }

  @override
  void dispose() {
    processing!.complete(false);
    _wrap!.dispose();
    super.dispose();
  }

  @override
  Widget build(BuildContext context) {
    final mq = MediaQuery.of(context);
    var crossCount = 3;
    var w = math.min(880, mq.size.width) / (mq.devicePixelRatio * crossCount);
    log.info("Width base is $w");
    // final rows = _mediaUris.chunked(widget.columns);
    var itemBuilder = (MediaSource<F?> item) {
      if (item is MediaSourcePlaceholder) {
        return MediaSourcePlaceholderWidget(
            size: widget.size, processing: processing, onTap: _uploadMedia);
      } else if (item is MediaSourceData<F>) {
        return MediaSourceDisplay(
          index: _mediaUris!.indexOf(item),
          discriminator: widget.name,
          media: item,
          size: widget.itemHeight ?? widget.size,
          onRemoved: () => _removeItem(item),
        );
      } else {
        return ErrorText("Invalid image");
      }
    };
    return ReorderableWrap(
      onReorder: (int oldIndex, int newIndex) {
        safeState(() {
          var moved = _mediaUris!.removeAt(oldIndex);
          _mediaUris!.insert(newIndex, moved);
          notifyCompleted();
        });
      },
      controller: _wrap,
      needsLongPressDraggable: false,
      maxMainAxisCount: widget.columns,
      spacing: sunnySpacing * 1,
      runSpacing: sunnySpacing * 1,
      children: [
        for (var r in _mediaUris!)
          SizedBox(
            key: Key("image-${r.diffKey}"),
            height: math.max(widget.itemHeight ?? 100, w),
            width: w,
            child: itemBuilder(r),
          ),
      ],
    );
  }

  _removeItem(MediaSource item) {
    mediaUris = [..._mediaUris!.where((u) => u != item)];
  }

  Future<Iterable<MediaSource>> _uploadMedia() async {
    if (widget.isReadOnly) return [];
    Hud? hud;

    final pickedMedia = await (showMediaListPicker(
      context,
      widget.type,
      onLoading: () {
        log.info("Started loading");
        hud = SunnyHud.spinner(context, "Reading...", duration: 1.minute);
      },
      imageSelector: Choosers.flutterFilePicker,
      maxSelections: widget.type.maxSelections,
      crop: !kIsWeb,
    ) as FutureOr<List<ChosenMedia>>);
    hud?.dismiss();
    log.info("Picked ${pickedMedia.length} items");

    final List<MediaSourceData<F?>> selectedFiles = [
      for (var p in pickedMedia)
        if (p.file != null)
          MediaSourceData<F>.ofPFile(
              widget.type, p.file, p.file!.name ?? uuid())
        else if (p.content != null)
          MediaSourceData<F?>.ofMediaContent(widget.type, p.content as F?,
              mediaId:
                  widget.type.mediaUri(p.content as F)?.toString() ?? uuid()),
    ];

    if (selectedFiles.isNotEmpty) {
      setState(() {
        /// Just clobber the old one
        processing!.complete();
        processing!.start();
      });
    }

    mediaUris = [..._mediaUris!, ...selectedFiles];
    try {
      final images =
          await (selectedFiles.mapIndexed((mediaSource, index) async {
        try {
          if (mediaSource.file != null) {
            assert(mediaSource.mediaId != null);

            final progress = mediaService.uploadMedia(
              mediaSource.file,
              widget.type,
              mediaId: mediaSource.mediaId,
              progress: mediaSource.progress,
            );

            final uri = await progress.result;
            mediaSource.progress?.complete();
            log.info("Completed uploading $uri");
            return mediaSource.uploaded(uri);
          } else {
            final uri = mediaSource.contentType
                .mediaUri(mediaSource.mediaContent)
                ?.toString();
            mediaSource.progress?.complete();
            processing!.complete();
            return mediaSource.uploaded(uri?.toUri());
          }
        } catch (e, stack) {
          log.severe("Error uploading single item: $e", e, stack);
          rethrow;
        }
      }).waitAll(eagerError: true) as FutureOr<List<MediaSourceData<F>>>);

      final replacement = [..._mediaUris!];
      for (var resolved in images) {
        final idx = _mediaUris!.indexWhere((final item) {
          return item == resolved.original;
        });
        replacement[idx] = resolved;
      }

      mediaUris = replacement;
      return replacement;
    } catch (e, stack) {
      widget.onError?.call(e, stack);
      rethrow;
    } finally {
      processing!.complete();
    }
  }
}
