import 'dart:async';

import 'package:collection_diff/collection_diff.dart';
import 'package:equatable/equatable.dart';
import 'package:flutter/cupertino.dart';
import 'package:flutter/foundation.dart';
import 'package:flutter/material.dart';
import 'package:flutter/widgets.dart';
import 'package:flutter_platform_widgets/flutter_platform_widgets.dart';
import 'package:pfile/pfile.dart';
import 'package:sunny_dart/helpers/safe_completer.dart';
import 'package:sunny_dart/sunny_dart.dart';
import 'package:sunny_forms/dialogs.dart';
import 'package:sunny_forms/dividing_line.dart';
import 'package:sunny_forms/media/media_content_type.dart';
import 'package:sunny_forms/state.dart';
import 'package:sunny_sdk_core/model/progress_tracker.dart';
import 'package:sunny_service_stubs/services.dart';

abstract class MediaSource<F> implements DiffDelegate {
  FutureOr get ready;
}

class MediaSourcePlaceholder<F> extends DiffDelegate implements MediaSource<F> {
  const MediaSourcePlaceholder() : super();

  @override
  get diffKey => "mediaSourcePlaceholder";

  @override
  get diffSource => "mediaSourcePlaceholder";

  void get ready {}
}

/// Could be an uploaded image (uri) or a file that the user chose that's in the process
/// of being uploaded
class MediaSourceData<F>
    with DiffDelegateMixin
    implements MediaSource<F>, Equatable {
  final F? mediaContent;
  final MediaContentType<F> contentType;

  /// ImageId is used for equality checks when uploading new files (because the
  /// file itself isn't very good to use in equality checks)
  final String? mediaId;

  /// The type of image
  final dynamic mediaCategory;

  /// If the file is uploading, this tracker will report the status
  final ProgressTracker<Uri>? progress;

  /// nullable: A file on the device
  final PFile? file;

  /// The original content
  final MediaSourceData<F>? original;

  MediaSourceData._({
    required this.mediaContent,
    required this.mediaId,
    required this.contentType,
    required this.mediaCategory,
    required this.progress,
    required this.file,
    required this.original,
  });

  MediaSourceData.ofMediaContent(this.contentType, this.mediaContent,
      {required String? mediaId})
      : mediaCategory = null,
        mediaId = mediaId ?? contentType.mediaUri(mediaContent!)?.toString(),
        progress = null,
        original = null,
        file = null;

  MediaSourceData.ofPlatformFile(this.contentType, file, this.mediaId,
      {this.mediaCategory})
      : progress = ProgressTracker.ratio(),
        file = PFile.of(file),
        mediaContent = null,
        original = null;

  MediaSourceData.ofPFile(this.contentType, this.file, this.mediaId,
      {this.mediaCategory})
      : progress = ProgressTracker.ratio(mediaId),
        original = null,
        mediaContent = null;

  MediaSourceData._uploaded(MediaSourceData<F> this.original, this.mediaContent)
      : file = original.file,
        contentType = original.contentType,
        progress = null,
        mediaId = null,
        mediaCategory = null;

  @override
  List<Object?> get props => [
        contentType.mediaUri(mediaContent!),
        mediaId,
        contentType.name,
        file?.path,
        original,
      ];

  @override
  bool get stringify => true;

  bool get isUploading {
    return progress != null && progress!.percent < 1;
  }

  bool get isCompleted => !isUploading;

  Future? get ready {
    if (file == null) {
      return SynchronousFuture(null);
    }
    return file?.read;
  }

  @override
  // Equals is overridden by Equatable - but we use the imageUrl hashcode
  // ignore: hash_and_equals
  int get hashCode => _key.hashCode;

  String? get uri => contentType.mediaUri(mediaContent!)?.toString();

  String? get _key => (mediaCategory != null || mediaId != null)
      ? [mediaCategory, mediaId].whereNotNull().join("/")
      : uri;

  @override
  get diffSource {
    var source = {"key": _key, isUploading: isUploading};
    print("diff source: $uri");
    return source;
  }

  @override
  String? get diffKey {
    print("diff: $_key");
    return _key;
  }

  MediaSourceData<F> uploaded(Uri? uri) {
    // ignore: missing_required_param
    return MediaSourceData<F>._uploaded(this, contentType.uriToMedia(uri));
  }

  MediaSourceData<F> copy({
    required F? mediaContent,
    required String? mediaId,
    required MediaContentType<F>? contentType,
    required mediaCategory,
    required ProgressTracker<Uri>? progress,
    required PFile? file,
    required MediaSourceData<F>? original,
  }) {
    return MediaSourceData._(
      mediaContent: mediaContent ?? this.mediaContent,
      mediaId: mediaId ?? this.mediaId,
      mediaCategory: mediaCategory ?? this.mediaCategory,
      contentType: contentType ?? this.contentType,
      progress: progress ?? this.progress,
      file: file ?? this.file,
      original: original ?? this.original,
    );
  }
}

/// The placeholder image (that the user taps to upload a new imageE)
class MediaSourcePlaceholderWidget extends StatelessWidget {
  final VoidCallback? onTap;
  final SafeCompleter? processing;
  final double size;

  const MediaSourcePlaceholderWidget(
      {Key? key, required this.size, required this.processing, this.onTap})
      : super(key: key);

  @override
  Widget build(BuildContext context) {
    return Container(
      alignment: AlignmentDirectional.center,
      height: 100,
      decoration: BoxDecoration(
        color: sunnyColors.scaffoldBackground,
        borderRadius: BorderRadius.all(const Radius.circular(10.0)),
      ),
      child: PlatformIconButton(
        icon: NamedIcons.upload.solid().sized(size).widget,
        onPressed: processing!.isCompleted ? onTap : null,
      ),
    );
  }
}

/// Displays a single image tile, which may be in the process of uploading
class MediaSourceDisplay extends StatelessWidget {
  /// The image to be displayed.
  final MediaSourceData media;

  final String discriminator;
  final int index;
  final double size;
  final bool isDeletable;
  final double borderRadius;
  final VoidCallback? onRemoved;

  const MediaSourceDisplay({
    Key? key,
    required this.discriminator,
    required this.media,
    this.isDeletable = true,
    required this.index,
    this.borderRadius = 12,
    required this.size,
    this.onRemoved,
  }) : super(key: key);

  @override
  Widget build(BuildContext context) {
    Widget w;
    if (media.isUploading == true) {
      w = media.contentType.buildPreview(
        context,
        media,
        child: Container(
          height: size,
          alignment: AlignmentDirectional.center,
          child: SunnyProgressIndicator(
            // key: Key("progress-${media.mediaId}"),
            tracker: media.progress,
            builder: (double) {
              var pct = double / media.file!.size;
              return Text("${(pct * 100).toInt()}%");
            },
            circular: false,
            showPercent: false,
          ),
        ),
        isUploading: true,
      );
    } else {
      w = Tappable(
        child: Hero(
          tag: "$discriminator-hero-$index-image",
          transitionOnUserGestures: true,
          child: media.contentType.buildThumbnail(
            context,
            media,
            isUploading: false,
            borderRadius: borderRadius,
          ),
        ),
        onTap: (context) => _openMedia(context),
        behavior: HitTestBehavior.translucent,
      );
    }

    if (isDeletable == true) {
      return Stack(alignment: Alignment.bottomCenter, children: [
        w,
        Container(
            height: 50,
            alignment: Alignment.bottomCenter,
            child: MaterialButton(
              child: NamedIcons.delete.build(color: sunnyColors.red),
              shape: CircleBorder(),
              onPressed: () async {
                if (await showConfirmDeleteDialog(context,
                    fullLabel:
                        "Do you want to remove this ${media.contentType.name.uncapitalize()}?")) {
                  onRemoved?.call();
                }
              },
            )),
      ]);
    } else {
      return w;
    }
  }

  Future _openMedia(BuildContext context) async {
    await Navigator.of(context).push(PageRouteBuilder(
      pageBuilder: (BuildContext context, Animation<double> animation,
              Animation<double> secondaryAnimation) =>
          GestureDetector(
        onTap: () {
          Navigator.pop(context);
        },
        child: Center(
          child: Layout.column().reset.build(
            [
              Hero(
                key: Key("overlay-hero-$index"),
                tag: "$discriminator-hero-$index-image",
                transitionOnUserGestures: true,
                child: ResponsiveConstraint(
                  maxWidth: 700,
                  child: Container(
                    constraints: BoxConstraints(minHeight: 700),
                    child: media.contentType
                        .buildPreview(context, media, isUploading: false),
                  ),
                ),
              ),
              if (isDeletable) ...[
                sunnySpacing.vspace(),
                PlatformIconButton(
                  onPressed: () async {
                    Navigator.pop(context);
                    await Future.delayed(300.ms);
                    onRemoved?.call();
                  },
                  color: sunnyColors.primaryColor,
                  icon:
                      NamedIcons.delete.colored(sunnyColors.red).sized(20).icon,
                )
              ]
            ],
          ),
        ),
      ),
      opaque: false,
      barrierColor: CupertinoColors.white,
      // barrierLabel: barrierLabel,
      maintainState: true,
      barrierDismissible: true,
    ));
  }
}
