import 'dart:math';
import 'dart:typed_data';
import 'dart:ui';

import 'package:crop/crop.dart';
import 'package:pfile_picker/pfile_picker.dart';
import 'package:flutter/cupertino.dart';
import 'package:flutter/material.dart';
import 'package:image/image.dart' as img;
import 'package:sunny_dart/helpers/strings.dart';
import 'package:sunny_dart/sunny_dart.dart';
import 'package:sunny_forms/dividing_line.dart';
import 'package:sunny_forms/platform/platform_flat_button.dart';
import 'package:sunny_forms/platform/platform_raised_button.dart';
import 'package:sunny_forms/sunny_hud.dart';

class CropImagePage extends StatefulWidget {
  final Uint8List source;

  const CropImagePage({Key? key, required this.source}) : super(key: key);
  CropImagePage.ofPlatformFile({
    Key? key,
    required PFile source,
    Uint8List? bytes,
    // ignore: deprecated_member_use
  })  : source = bytes ?? source.bytes ?? illegalState("Need loaded bytes"),
        super(key: key);

  @override
  _CropImagePageState createState() => _CropImagePageState();
}

class _CropImagePageState extends State<CropImagePage> {
  CropController? controller;
  bool _isSubmitting = false;
  late Size _boxSize;
  @override
  void initState() {
    super.initState();
    controller = CropController();
  }

  @override
  void dispose() {
    controller!.dispose();
    super.dispose();
  }

  @override
  Widget build(BuildContext context) {
    return Column(
      children: <Widget>[
        Expanded(
          child: Container(
            color: Colors.black,
            padding: EdgeInsets.all(8),
            child: LayoutBuilder(
              builder: (_, layout) {
                _boxSize = layout.biggest;
                print("Crop box: $_boxSize");
                return Crop(
                  onChanged: (decomposition) {
                    // print(
                    //     "Scale : ${decomposition.scale}, Rotation: ${decomposition.rotation}, translation: ${decomposition.translation}");
                  },
                  dimColor: Colors.transparent,
                  controller: controller!,
                  shape: BoxShape.rectangle,
                  child: Image(
                    image: MemoryImage(widget.source),
                    fit: BoxFit.cover,
                  ),
                  /* It's very important to set `fit: BoxFit.cover`.
                     Do NOT remove this line.
                     There are a lot of issues on github repo by people who remove this line and their image is not shown correctly.
                  */
                  helper: Container(
                    decoration: BoxDecoration(
                      border: Border.all(
                        color: Colors.white,
                        width: 2,
                      ),
                    ),
                  ),
                );
              },
            ),
          ),
        ),
        CropWidgetControls(
          controller: controller,
        ),
        SizedBox(
          height: 50,
          child: Layout.row().alignCenter.build(
                PlatformRaisedButton(
                  child: textOrNull(_isSubmitting ? "Processing..." : "Save",
                      color: Colors.white),
                  color: sunnyColors.primaryColor,
                  onPressed: () async {
                    setState(() {
                      _isSubmitting = true;
                    });

                    try {
                      Future.microtask(
                        () async {
                          // Resize and pop
                          final pixelRatio =
                              MediaQuery.of(context).devicePixelRatio;

                          final scale = controller!.scale;
                          final uiOffset = controller!.offset;

                          print("Pixel Ratio: $pixelRatio");
                          print("Scale: $scale");
                          print("Offset: $uiOffset");

                          var decoded = img.decodeImage(widget.source)!;
                          final imgWidth = decoded.width.toDouble();
                          final imgHeight = decoded.height.toDouble();

                          final box = min(_boxSize.height, _boxSize.width);
                          final baseScale = min(imgWidth, imgHeight);
                          final ratio = box / baseScale;

                          final offset = uiOffset / ratio;

                          final cropArea = Size(baseScale, baseScale);

                          final outerW = imgWidth * scale;
                          final outerH = imgHeight * scale;

                          double pXa = 0.5 * (outerW - cropArea.width);
                          double pYa = 0.5 * (outerH - cropArea.height);

                          // final isPortrait = imgWidth < imgHeight;
                          //
                          // /// Aspect multipliers - depending on whether it's
                          // /// landscape or portrait mode
                          // final wx = isPortrait ? 1 : imgHeight / imgWidth;
                          // final hx = !isPortrait ? 1 : imgWidth / imgHeight;
                          //
                          // /// Calculate initial shifts required for aspect
                          // final wshift = isPortrait ? 0 : wx / 2;
                          // final hshift = isPortrait ? 0 : hx / 2;

                          double xPercent = pXa != 0.0
                              ? 1.0 - (offset.dx + pXa) / (2 * pXa)
                              : 0.0;
                          double yPercent = pYa != 0.0
                              ? 1.0 - (offset.dy + pYa) / (2 * pYa)
                              : 0.0;
                          double cropXpx = imgWidth * cropArea.width / outerW;
                          double cropYpx = imgHeight * cropArea.height / outerH;
                          double x0 = (imgWidth - cropXpx) * xPercent;
                          double y0 = (imgHeight - cropYpx) * yPercent;
                          final cropped = img.copyCrop(decoded, x0.toInt(),
                              y0.toInt(), cropXpx.toInt(), cropYpx.toInt());
                          final encoded = await img.encodeJpg(cropped);

                          final uid = "${uuid()}.jpg";

                          Navigator.pop(
                              context,
                              PlatformFile(
                                path: 'tmp/$uid',
                                name: uid,
                                bytes: Uint8List.fromList(encoded),
                                size: encoded.length,
                              ));
                        },
                      );
                    } catch (e) {
                      SunnyHud.error(context, "Error: $e");
                      Navigator.pop(context);
                    }
                  },
                ),
                sunnySpacing.horizSpace,
                PlatformFlatButton(
                  child: Text("Cancel"),
                  onPressed: () {
                    if (mounted) {
                      Navigator.of(context).pop();
                    }
                  },
                ),
              ),
        ),
      ],
    );
  }
}

class CropWidgetControls extends StatefulWidget {
  final CropController? controller;

  const CropWidgetControls({Key? key, required this.controller})
      : super(key: key);

  @override
  _CropWidgetControlsState createState() => _CropWidgetControlsState();
}

class _CropWidgetControlsState extends State<CropWidgetControls> {
  double _slider = 1.0;

  @override
  Widget build(BuildContext context) {
    return SizedBox(
      height: 50,
      child: Row(
        children: <Widget>[
          IconButton(
            icon: Icon(Icons.undo),
            tooltip: 'Undo',
            onPressed: () {
              widget.controller!.rotation = 0;
              widget.controller!.scale = 1;
              widget.controller!.offset = Offset.zero;
              setState(() {
                _slider = 1.0;
              });
            },
          ),
          Expanded(
            child: Slider.adaptive(
              divisions: 100,
              value: _slider,
              min: 1.0,
              max: 3.0,
              label: '$_slider°',
              onChanged: (n) {
                _slider = n;
                setState(() {
                  _slider = n;
                });
                print("Setting scale: $_slider");
                widget.controller!.scale = _slider;
              },
            ),
          ),
        ],
      ),
    );
  }
}
