import 'package:chewie/chewie.dart';
import 'package:equatable/equatable.dart';
import 'package:flutter/foundation.dart';
import 'package:flutter/material.dart';
import 'package:flutter/widgets.dart';
import 'package:muse_ai/muse_ai.dart';
import 'package:sunny_dart/sunny_dart.dart';
import 'package:sunny_forms/dividing_line.dart';
import 'package:sunny_forms/form_fields/media_source.dart';
import 'package:sunny_forms/icon/image_widgets.dart';
import 'package:sunny_forms/muse.dart';
import 'package:sunny_forms/players/stateful_video_player.dart';
import 'package:sunny_service_stubs/models.dart';

/// Represents either image or video content, allows for different upload/preview
abstract class MediaContentType<F> extends Equatable {
  const MediaContentType();

  String get name;

  Color get barrierColor;

  int get maxSelections;

  JsonPath get fieldPath;

  Uri? mediaUri(F fact);

  F uriToMedia(Uri? uri);

  List<String> get acceptedFileTypes;

  @override
  List<Object> get props => [name];

  Widget buildPreview(BuildContext context, MediaSourceData source,
      {Widget? child, required bool isUploading, double? borderRadius});

  Widget buildThumbnail(
    BuildContext context,
    MediaSourceData source, {
    Widget? child,
    required bool isUploading,
    double? borderRadius,
  });

  F newContent();

  static VideoContentType get video => const VideoContentType._();

  static ImageContentType get image => const ImageContentType._();
}

class VideoContentType extends MediaContentType<IVideoContent?> {
  const VideoContentType._();

  @override
  List<String> get acceptedFileTypes => ["mov", "mp4"];

  @override
  Uri? mediaUri(IVideoContent? fact) => fact?.videoUrl?.toUri();

  @override
  String get name => "video";

  @override
  IVideoContent? uriToMedia(Uri? uri) {
    return uri == null
        ? null
        : IVideoContent.of(
            videoUrl: uri.toString(),
            autoplay: false,
            showControls: true,
            dismissible: false,
            delivery: IContentDelivery.internal("inline"),
          );
  }

  @override
  Widget buildPreview(BuildContext context, MediaSourceData media,
      {Widget? child, required bool isUploading, double? borderRadius}) {
    final player = buildVideo(
      context,
      media,
      child: child,
      isUploading: isUploading,
      borderRadius: borderRadius,
    );
    return Stack(children: [
      Container(
        decoration: BoxDecoration(
          color: sunnyColors.g100,
          borderRadius: BorderRadius.all(Radius.circular(borderRadius ?? 10)),
        ),
      ),
      player,
      if (child != null && isUploading) child,
    ]);
  }

  Widget buildVideo(BuildContext context, MediaSourceData media,
      {Widget? child,
      bool? isUploading,
      double? borderRadius,
      Consumer<ChewieController>? controller}) {
    ChewieController? _chewie;
    Widget player;
    if (media.uri!.startsWith("https://muse.ai")) {
      try {
        final museIndex = sunny.get<MuseIndex>();
        final args = MuseArgs.ofUri(media.uri!);
        final video = museIndex.findVideoBySvid(args.museId)!;
        player = StatefulVideoPlayer(
          key: Key("videoUrl${media.uri}"),
          videoUrl: video.format("mp4"),
          thumbnailUrl: null,
          // isThumbnail: true,
          showControls: true,
          autoplay: false,
          initializer: (chewie) {
            _chewie ??= chewie;
            controller?.call(chewie!);
            if (args.start > 0) {
              chewie!.videoPlayerController
                  .seekTo(Duration(seconds: args.start));
            }
          },
          onCompletion: () {},
        );
      } catch (e) {
        rethrow;
      }
    } else {
      player = media.uri != null
          ? StatefulVideoPlayer(
              key: Key("videoUrl${media.uri}"),
              videoUrl: media.uri,
              thumbnailUrl: null,
              // isThumbnail: true,
              showControls: true,
              autoplay: false,
              initializer: (chewie) {
                _chewie ??= chewie;
              },
              onCompletion: () {},
            )
          : StatefulVideoPlayer(
              key: Key("videoFile${media.file?.path ?? media.file?.name}"),
              videoFile: media.file,
              // isThumbnail: true,
              thumbnailUrl: null,
              autoplay: false,
              showControls: true,
              initializer: (chewie) {
                _chewie ??= chewie;
              });
    }
    return player;
  }

  @override
  Widget buildThumbnail(BuildContext context, MediaSourceData media,
      {Widget? child, bool? isUploading, double? borderRadius}) {
    ChewieController? chewie;
    final player = buildVideo(
      context,
      media,
      child: child,
      isUploading: isUploading,
      borderRadius: borderRadius,
      controller: (_) => chewie = _,
    );
    return Tappable(
      onTap: (_) async {
        chewie?.enterFullScreen();
        chewie?.play();
      },
      child: Stack(
        alignment: Alignment.center,
        children: [
          Container(
            decoration: BoxDecoration(
              color: sunnyColors.g100,
              borderRadius:
                  BorderRadius.all(Radius.circular(borderRadius ?? 10)),
            ),
          ),
          player,
          if (child != null && isUploading!) child,
        ],
      ),
    );
  }

  @override
  Color get barrierColor => Colors.white;

  @override
  int get maxSelections => 1;

  @override
  IVideoContent newContent() {
    // ignore: missing_required_param
    return IVideoContent.of(
      autoplay: false,
      showControls: false,
      videoUrl: null,
      dismissible: false,
      delivery: IContentDelivery.inline_,
    );
  }

  @override
  JsonPath get fieldPath => VideoContentPaths.videoUrl;
}

class ImageContentType extends MediaContentType<IImageContent?> {
  const ImageContentType._();

  @override
  List<String> get acceptedFileTypes => ["jpg", "jpeg", "gif", "png"];

  @override
  Uri? mediaUri(IImageContent? fact) => fact?.imageUrl?.toUri();

  @override
  String get name => "image";

  @override
  JsonPath get fieldPath => ImageContentPaths.imageUrl;

  @override
  IImageContent? uriToMedia(Uri? uri) {
    return uri == null
        ? null
        : IImageContent.of(
            imageUrl: uri.toString(),
            delivery: IContentDelivery.inline_,
            dismissible: false);
  }

  @override
  IImageContent newContent() {
    // ignore: missing_required_param
    return IImageContent.of(
      dismissible: false,
      imageUrl: null,
      delivery: IContentDelivery.inline_,
    );
  }

  @override
  Widget buildThumbnail(BuildContext context, MediaSourceData source,
      {Widget? child, required bool isUploading, double? borderRadius}) {
    return buildPreview(context, source,
        child: child, isUploading: isUploading, borderRadius: borderRadius);
  }

  @override
  Widget buildPreview(BuildContext context, MediaSourceData media,
      {Widget? child, required bool isUploading, double? borderRadius}) {
    if (isUploading) {
      return Container(
        decoration: BoxDecoration(
          image: DecorationImage(
            image: imageProviderOf(media.file)!,
            colorFilter: ColorFilter.mode(
                Colors.white.withOpacity(0.6), BlendMode.lighten),
            fit: BoxFit.cover,
          ),
          borderRadius: BorderRadius.all(Radius.circular(borderRadius ?? 10)),
        ),
        child: child,
      );
    } else {
      return Container(
        decoration: BoxDecoration(
          image: media.uri == null
              ? null
              : DecorationImage(
                  image: imageProviderOf(media.uri)!,
                  fit: BoxFit.cover,
                ),
          borderRadius: BorderRadius.all(Radius.circular(borderRadius ?? 10.0)),
        ),
      );
    }
  }

  @override
  Color get barrierColor => sunnyColors.barrierColor;

  @override
  int get maxSelections => 1;
}

extension MCTExt on MediaContentType? {
  bool get isImage {
    return this == MediaContentType.image;
  }

  bool get isVideo {
    return this == MediaContentType.video;
  }
}
