import 'package:meta_forms/meta_forms.dart';
import 'package:sunny_dart/json/json_path.dart';
import 'package:sunny_dart/sunny_dart.dart';
import 'package:sunny_sdk_core/mverse/m_model.dart';
import 'package:sunny_sdk_core/sunny_sdk_core.dart';

typedef PropertyBuilder = void Function(MSchemaPropertyBuilder prop);

/// Convenient way to build entity schemas on the client.  Typically, entity schemas come from the mverse server, but
/// sometimes you want to build a metaform because it's easier than building a form manually.
class EntitySchemas {
  static IMEntityDefinition build(
      MSchemaRef schemaRef, void builder(EntitySchemas schema)) {
    final schema = EntitySchemas._(schemaRef);
    builder(schema);
    return schema._build();
  }

  final MSchemaRef schemaRef;
  final Uri? schemaUri;
  final Map<JsonPath, IMSchemaProperty> properties = {};
  MSchemaRef? parentRef;

  void _addProperty<T extends MSchemaPropertyBuilder>(
      path, T target, dynamic builder(T builder)) {
    final jsonPath = JsonPath.of(path);
    builder(target);
    properties[jsonPath] = target.build();
  }

  void prop(path, dynamic build(MPropertyBuilder builder)) =>
      _addProperty(path, MPropertyBuilder(schemaUri, JsonPath.of(path)), build);

  void embeddedProp(path, dynamic build(EmbeddedPropertyBuilder builder)) =>
      _addProperty(
          path, EmbeddedPropertyBuilder(schemaUri, JsonPath.of(path)), build);

  void entityListProp(path, dynamic build(EntityListPropertyBuilder builder)) =>
      _addProperty(
          path, EntityListPropertyBuilder(schemaUri, JsonPath.of(path)), build);

  void refProp(path, dynamic build(RefPropertyBuilder builder)) => _addProperty(
      path, RefPropertyBuilder(schemaUri, JsonPath.of(path)), build);

  void stringSelectProp(
          path, dynamic build(StringSelectPropertyBuilder builder)) =>
      _addProperty(path,
          StringSelectPropertyBuilder(schemaUri, JsonPath.of(path)), build);

  EntitySchemas._(this.schemaRef)
      : schemaUri = schemaRef.relativeSchemaUri.toUri();

  IMEntityDefinition _build() {
    return IMEntityDefinition.of(
      schemaURI: schemaUri,
      properties: this.properties.values.toList(),
      schemaRef: schemaRef,
      parentRef: parentRef,
    );
  }
}

abstract class MSchemaPropertyBuilder {
  MSchemaPropertyBuilder(this.schemaUri, this.path);

  /// Will use [definition.propType] if it exists
  PropSchemaType get propType;

  /// The uri of the containing schema
  final Uri? schemaUri;
  final JsonPath path;
  bool isRequired = false;
  String? label;
  String? description;

  Definition? definition;

  IMSchemaProperty build();

  // _verify(String debug, bool expr()) {
  //   if (!expr()) {
  //     throw "Illegal property definition: $debug";
  //   }
  // }
}

class EmbeddedPropertyBuilder extends MSchemaPropertyBuilder {
  EmbeddedPropertyBuilder(Uri? schemaUri, JsonPath path)
      : super(schemaUri, path);

  MSchemaRef? embedRef;

  @override
  IMSchemaProperty build() {
    assert(embedRef != null);
    // ignore: missing_required_param
    return IEmbeddedProperty.of(
            path: null,
            type: null,
            isRequired: false,
            label: null,
            embedRef: embedRef) +
        this;
  }

  @override
  PropSchemaType get propType => PropSchemaType.embedded;
}

class EntityListPropertyBuilder extends MSchemaPropertyBuilder {
  EntityListPropertyBuilder(Uri? schemaUri, JsonPath path)
      : super(schemaUri, path);

  MSchemaRef? allItemsRef;

  @override
  IMSchemaProperty build() {
    assert(allItemsRef != null);
    return IEntityListProperty.of(
            type: null,
            path: null,
            label: null,
            isRequired: false,
            mkeyOnly: null,
            allItemsRef: allItemsRef) +
        this;
  }

  @override
  PropSchemaType get propType => PropSchemaType.entityList;
}

class RefPropertyBuilder extends MSchemaPropertyBuilder {
  RefPropertyBuilder(Uri? schemaUri, JsonPath path) : super(schemaUri, path);

  MSchemaRef? theRef;
  bool mkeyOnly = true;

  @override
  IMSchemaProperty build() {
    assert(theRef != null);
    return IRefProperty.of(
            type: null,
            path: null,
            label: null,
            isRequired: false,
            theRef: theRef,
            mkeyOnly: mkeyOnly) +
        this;
  }

  @override
  PropSchemaType get propType {
    return PropSchemaType.ref;
  }
}

class StringSelectPropertyBuilder extends MSchemaPropertyBuilder {
  StringSelectPropertyBuilder(Uri? schemaUri, JsonPath path)
      : super(schemaUri, path);

  List<String> options = [];

  @override
  IMSchemaProperty build() {
    assert(options.isNotNullOrEmpty);
    return IStringSelectProperty.of(
            type: null,
            path: null,
            label: null,
            isRequired: false,
            options: options) +
        this;
  }

  @override
  PropSchemaType get propType {
    return PropSchemaType.stringSelect;
  }
}

class MPropertyBuilder extends MSchemaPropertyBuilder {
  MPropertyBuilder(Uri? schemaUri, JsonPath path) : super(schemaUri, path);

  PropSchemaType _propType = PropSchemaType.string;

  set propType(PropSchemaType propType) {
    this._propType = propType;
  }

  @override
  PropSchemaType get propType {
    return definition?.propType ?? _propType;
  }

  @override
  IMSchemaProperty build() {
    final combined = IMSchemaProperty.of(
          type: null,
          path: null,
          label: null,
          isRequired: false,
        ) +
        this;
    return combined;
  }
}

enum MSchemaPropertyType {
  embeddedProperty,
  entityList,
  ref,
  property,
  stringSelect
}

extension MSchemaPropertyExtensions on IMSchemaProperty {
  IMSchemaProperty operator +(MSchemaPropertyBuilder builder) {
    final prop = this;
    assert(builder.path != JsonPath.root(), "path is not root");
    prop.type = builder.propType.value;
    prop.path = "${builder.path}";
    prop.label = builder.label ?? builder.path.toKey().toTitle();
    prop.description = builder.description;
    prop.uri = Uri.parse("${builder.schemaUri}${builder.path}");
    prop.baseCode = builder.definition?.baseCode;
    prop.uri = builder.definition?.uri.toUri();
    prop.isRequired = builder.isRequired;
    return prop;
  }
}
