import 'package:flutter/material.dart';
import 'package:meta_forms/meta_forms.dart';
import 'package:meta_forms/mschema_service.dart';
import 'package:sunny_dart/helpers/logging_mixin.dart';
import 'package:sunny_dart/json.dart';
import 'package:sunny_dart/sunny_dart.dart';
import 'package:sunny_forms/dividing_line.dart';
import 'package:sunny_forms/field_status.dart';
import 'package:sunny_forms/meta/props/collapsing_embed_property_handler.dart';
import 'package:sunny_forms/meta/props/default_property_handler.dart';
import 'package:sunny_forms/meta/props/duration_property_handler.dart';
import 'package:sunny_forms/meta/props/image_gallery_property_handler.dart';
import 'package:sunny_forms/meta/props/image_upload_property_handler.dart';
import 'package:sunny_forms/meta/props/location_property_handler.dart';
import 'package:sunny_forms/meta/props/mverse_embed_option_handler.dart';
import 'package:sunny_forms/meta/props/mverse_link_option_handler.dart';
import 'package:sunny_forms/meta/props/phone_with_extension_property_handler.dart';
import 'package:sunny_forms/meta/props/range_property_handler.dart';
import 'package:sunny_forms/meta/props/video_gallery_property_handler.dart';
import 'package:sunny_forms/options/options_service.dart';
import 'package:sunny_forms/options/options_view_service.dart';
import 'package:sunny_forms/sunny_form_controller.dart';
import 'package:sunny_sdk_core/mverse/definitions.dart';
import 'package:sunny_sdk_core/mverse/m_model.dart';
import 'package:sunny_sdk_core/mverse/mmodel_registry.dart';
import 'package:sunny_service_stubs/meta.dart';
import 'package:sunny_service_stubs/options.dart';

import 'meta_field_config.dart';
import 'meta_form.dart';
import 'meta_form_context.dart';
import 'meta_property_handler.dart';

MetaFormService? _metaFormService;

MetaFormService get metaFormService =>
    _metaFormService ?? illegalState("Must initialize global MetaFormService");

set metaFormService(MetaFormService _) {
  _metaFormService = _;
}

const contactKey = JsonPath.internal(["contactKey"], "/contactKey");
const contactKeys = JsonPath.internal(["contactKeys"], "/contactKeys");

final globalLabels = {contactKey: "Contact", contactKeys: "Contacts"};

extension MetaFormServiceBuildContextExt on BuildContext {
  MetaFormService get metaFormService =>
      (_metaFormService ??= Provided.get(this))!;
}

class MetaFormService with LoggingMixin implements IMetaFormRenderService {
  final OptionsService optionService;
  final TypeaheadService optionViewService;
  MetaFormService(
    MSchemaService schemaService,
    //    ContactMKeyOptionsHandler handler,
    this.optionService,
    this.optionViewService,
  ) : _globalOverrides = MetadataOverrides(
          labels: {contactKey: "Contact", contactKeys: "Contacts"},
          placeholders: {
            contactKey: "Choose a contact...",
            contactKeys: "Choose contacts"
          },
        ) {
    _metaFormService = this;
    handlers.add(PhoneWithExtensionPropertyHandler());
    handlers.add(DurationPropertyHandler());
    // handlers.add(ContactLinkOptionHandler(handler));
    handlers.add(ImageUploadPropertyHandler());
    handlers.add(ImageGalleryPropertyHandler());
    handlers.add(VideoGalleryPropertyHandler());
    handlers.add(MverseLinkOptionHandler());
    handlers.add(MverseEmbedOptionHandler());
    handlers.add(LocationPropertyHandler(schemaService));
    handlers.add(RangePropertyHandler());
    // handlers.add(ReachOutTypePropertyHandler());
    handlers.add(CollapsingEmbedPropertyHandler({}));
    handlers.add(DefaultPropertyHandler(optionService, optionViewService));
  }

  final Map<MSchemaRef, MetaFormBuilder> _formBuilders = {};
  final handlers = <MetaPropertyHandler>[];
  MetadataOverrides _globalOverrides;

  void addGlobalOverrides(MetadataOverrides overrides) {
    _globalOverrides += overrides;
  }

  void addOverrides(MSchemaRef ref, MetadataOverrides overrides) {
    assert(!_overrides.containsKey(ref), "Already have overrides for $ref");

    _overrides[ref] = overrides;
  }

  final _overrides = <MSchemaRef, MetadataOverrides>{};

  // static const _textAreaExpanding = TextAreaConfig(
  //   minLines: 3,
  //   maxLines: 18,
  // );

  MetaFormService operator +(MetaPropertyHandler handler) {
    handlers.add(handler);
    return this;
  }

  HandledPaths processSingle(
    IMSchemaDefinition definition, {
    JsonPath? basePath,
    required JsonPath singlePath,
    Iterable<MetaPropertyHandler>? adhocHandlers,
  }) {
    basePath ??= JsonPath.root();
    final result = MetaFormResult.single(
        definition: definition, basePath: basePath, single: singlePath);
    return result.handled[singlePath] ??
        nullPointer("No handled path for $singlePath");
  }

  /// Examines a schema definition, and locates the appropriate handlers for each logical property.  The [basePath] is
  /// used in the case where a subform is "flattened" into it's parent, and helps the system recognize and read/store
  /// data to the subtree, even though the visual representation is flattened.
  MetaFormResult process(
    IMSchemaDefinition definition, {
    JsonPath? basePath,
    Iterable<IMetaPropertyHandler>? adhocHandlers,
  }) {
    basePath ??= JsonPath.root();
    final result = MetaFormResult(definition: definition, basePath: basePath);
    [...?adhocHandlers, ...handlers].forEach((handler) {
      final accepted = [
        ...(handler as MetaPropertyHandler)
            .acceptProperties(basePath, definition, result.pending)
      ];
      for (final handled in accepted) {
        // Verify that these fields aren't already handled
        final alreadyMapped = handled.paths.keys
            .where((path) => result.mappedFields.containsKey(path));
        if (alreadyMapped.isNotEmpty) {
          log.info(
              "Skipping paths ${handled.paths} because the following were already mapped: $alreadyMapped");
          continue;
        }

        handled.paths.keys.forEach((path) {
          result.mappedFields[path] = handled;
          result.pending.remove(path);
        });

        result.handled[handled.path] = handled;
      }
    });

    return result;
  }

  Future<MetaFormContext> createContext(
    BuildContext context,
    MModel? subject, {
    MSchemaRef? schemaRef,
    MetadataOverrides? overrides,
    IMSchemaDefinition? schema,
    JsonPath? embedPath,
    Map<String, dynamic>? scope,
    List<MetaPropertyHandler> extraHandlers = const [],
  }) async {
    schemaRef ??= subject?.mtype;
    schema ??= (await context.mschemaService.getSchema(schemaRef));

    return MetaFormContext(context,
        subject: subject,
        schema: schema,
        ref: schemaRef!,
        scope: scope,
        embedPath: embedPath,
        overrides: getOverrides(schemaRef) + overrides,
        extraHandlers: extraHandlers);
  }

  MetadataOverrides getOverrides(MSchemaRef schemaRef) {
    var thisSchemaOverrides = _overrides[schemaRef];
    return _globalOverrides +
        (thisSchemaOverrides ?? const MetadataOverrides.empty());
  }

  @override
  Widget buildForm(BuildContext context, MSchemaRef ref,
      {final IFormController? controller,
      MetadataOverrides? overrides,
      MModel? record}) {
    if (controller != null && controller is MModelForm) {
      var formRef = ref;
      return MetaForm(
        context: context,
        parentForm: controller,
        attachForm: true,
        inheritForm: true,
        skipIfUnmodified: false,
        name: "$controller",
        ref: formRef,
        metaFormContext: metaFormService.createContext(
          context,
          controller.mmodel,
          schemaRef: formRef,
          overrides: overrides,
          // extraHandlers: extraHandlers,
          // embedPath: embedPath,
          // schema: schema,
          scope: {},
        ),
      );
    } else {
      record ??=
          mmodelRegistry.instantiate(json: <String, dynamic>{}, type: ref);
      final override = _formBuilders[ref] ??
          (context, ref, record) =>
              defaultMModelFormBuilder(context, ref, record);

      final builtForm = override(context, ref, record);
      return builtForm;
    }
  }

  Widget defaultMModelFormBuilder(
      BuildContext context, MSchemaRef ref, MModel mmodel) {
    final showContact = false;

    return mschemaService.getSchema(ref).build((schema) {
      return MetaForm.ofMModel(
        context,
        model: mmodel,
        saveModel: true,
        attachForm: true,
        overrides: MetadataOverrides(
          ignoredPaths: {
            if (showContact == FieldStatus.hidden) JsonPath.of("/contactKey")
          },
          formConfig: {
            JsonPath.of("/contactKey"): [
              MetaFieldConfig.of(
                  isDisabled: showContact == FieldStatus.readOnly)
            ],
          },
        )
        // + overrides
        ,
        schema: schema,
        skipIfUnmodified: false,
      );
    }, crossFade: false);
  }

// MetaDateFormatter getMetaDateOverrides(
//     MSchemaRef factType, FactMetaDate metaDate) {
//   return getOverrides(factType).metaDateFormatters[metaDate.path] ??
//       MetaDateFormatter.fallback();
// }
}

/// Results of processing a schema definition for rendering
class MetaFormResult {
  /// The schema being processed
  final IMSchemaDefinition definition;

  /// A base path to append
  final JsonPath basePath;

  /// The fields that have not been processed yet
  final Map<JsonPath, IMSchemaProperty> pending;

  /// Any fields that have been mapped already
  final mappedFields = <JsonPath, HandledPaths>{};

  /// Handled paths - the key in this map is [HandledPaths.path]
  final handled = <JsonPath, HandledPaths>{};

  HandledPaths? operator [](JsonPath path) {
    return handled[path];
  }

  MetaFormResult({required this.definition, this.basePath = JsonPath.Root})
      : pending = mapOf(definition.properties!, keyOf: (prop) => prop.jsonPath);

  MetaFormResult.single(
      {required this.definition,
      this.basePath = JsonPath.Root,
      required JsonPath single})
      : pending = {
          single: definition.properties!
                  .where((prop) => prop.jsonPath == single)
                  .firstOrNull ??
              nullPointer(
                  "No single result found for $single (base $basePath), but found "
                  "${definition.properties!.map(
                (p) => p.path,
              )}")
        };

  /// Gets a handled path.  The basePath is what the caller thinks it's root path.  All the paths in [handled] are
  /// intended to be relative, so, we'll ensure that the basePath matches, to prevent returning a false positive.
  HandledPaths? get(JsonPath? compareBasePath, JsonPath propertyPath) {
    compareBasePath ??= JsonPath.root();
    return this.basePath == compareBasePath ? handled[propertyPath] : null;
  }
}

/// Overrides for meta date properties
//abstract class MetaDateFormatter<T> {
//  String title(T fact, Contact contact);
//
//  String subtitle(T fact, Contact contact);
//
//  Icon icon(T fact, Contact contact, {double size = kTileIconSize});
//
//  factory MetaDateFormatter.of({
//    @required String Function(T fact, Contact contact) title,
//    @required String Function(T fact, Contact contact) subtitle,
//    @required Icon Function(T fact, Contact contact, {double size}) icon,
//  }) =>
//      _MetaDateFormatter<T>(title: title, subtitle: subtitle, icon: icon);
//
//  static MetaDateFormatter<T> fallback<T extends Fact>() =>
//      _MetaDateFormatter<T>(
//        title: (T fact, Contact contact) => fact.title(contact),
//        subtitle: (T fact, Contact contact) => fact.subtitle(contact),
//        icon: (T fact, Contact contact, {double size}) =>
//            SunnyIcons.getIconOrNull(fact.factSchema.icon)?.sized(size)?.icon,
//      );
//}

//class _MetaDateFormatter<T> implements MetaDateFormatter<T> {
//  final String Function(T fact, Contact contact) _title;
//  final String Function(T fact, Contact contact) _subtitle;
//  final Icon Function(T fact, Contact contact, {double size}) _icon;
//
//  @override
//  Icon icon(T fact, Contact contact, {double size = kTileIconSize}) {
//    return _icon(fact, contact, size: size);
//  }
//
//  @override
//  String subtitle(T fact, Contact contact) {
//    return _subtitle(fact, contact);
//  }
//
//  @override
//  String title(T fact, Contact contact) {
//    return _title(fact, contact);
//  }
//
//  _MetaDateFormatter({
//    @required String Function(T fact, Contact contact) title,
//    @required String Function(T fact, Contact contact) subtitle,
//    @required Icon Function(T fact, Contact contact, {double size}) icon,
//  })  : assert(title != null),
//        assert(subtitle != null),
//        assert(icon != null),
//        _title = title,
//        _subtitle = subtitle,
//        _icon = icon;
//}

final textFieldDefinitionOverrides = {
  Definitions.givenName:
      TextConfig(textCapitalization: TextCapitalization.words),
  Definitions.familyName:
      TextConfig(textCapitalization: TextCapitalization.words),
  Definitions.email: TextConfig(
      textInputType: TextInputType.emailAddress,
      textCapitalization: TextCapitalization.none),
  Definitions.phone: TextConfig(
      textInputType: TextInputType.phone,
      textCapitalization: TextCapitalization.none),
  Definitions.uri: TextConfig(textInputType: TextInputType.url),
  Definitions.interest: TextConfig(textCapitalization: TextCapitalization.none),
  Definitions.titleField:
      TextConfig(textCapitalization: TextCapitalization.sentences),
  Definitions.variableName:
      TextConfig(textCapitalization: TextCapitalization.none),
  Definitions.liquidContent:
      TextConfig(textCapitalization: TextCapitalization.none),
};

final textFieldOverridesByUri =
    textFieldDefinitionOverrides.mapKeys((key, _) => key.uri);
