import 'dart:async';

import 'package:flutter/material.dart';
import 'package:info_x/info_x.dart';
import 'package:logging/logging.dart';
import 'package:sunny_core_widgets/intl/sunny_intl.dart';
import 'package:sunny_dart/json.dart';
import 'package:sunny_dart/sunny_dart.dart';
import 'package:sunny_dart/typedefs.dart';
import 'package:sunny_essentials/sunny_essentials.dart';
import 'package:sunny_forms/icon/icon_widgets.dart';
import 'package:sunny_forms/options/mixins.dart';
import 'package:sunny_forms/platform/platform_bottom_sheet.dart';
import 'package:sunny_forms/platform/platform_bottom_sheet_action.dart';
import 'package:sunny_sdk_core/mverse/entity_extensions.dart';
import 'package:sunny_sdk_core/mverse/m_enum_registry.dart';
import 'package:sunny_sdk_core/services/sunny.dart';
import 'package:sunny_service_stubs/sunny_service_stubs.dart';

final _log = Logger("optionsService");

typedef AdhocBuilder<T> = AdhocOption<T> Function(String message);

// OptionsService get optionsService => Sunny.get();

@deprecated
Future<KeyedOption<K, V>?> openPicker<K, V>(BuildContext context,
        {title, content, required Set<KeyedOption<K, V>> options}) =>
    showPicker(context, title: title, content: content, options: options);

Future<KeyedOption<K, V>?> showPicker<K, V>(BuildContext context,
    {title, content, required Set<KeyedOption<K, V>> options}) async {
  final intl = SunnyIntl.of(context);
  options = options.toSet();
  return await showPlatformBottomSheet<KeyedOption<K, V>>(
      context: context,
      builder: (context) => PlatformBottomSheet(
            title: textOrNull(title),
            content: textOrNull(content),
            actions: [
              ...options.map(
                (KeyedOption<K, V> option) {
                  return PlatformBottomSheetAction(

                    child: PlatformListTile(
                      color: Colors.transparent,
                      onTap: (_)=>Navigator.pop(context, option),
                      padding: EdgeInsets.zero,
                      margin: EdgeInsets.zero,
                      title: textOrNull(option.label),
                      leading: (option.icon is IconData)
                          ? iconOrNull(option.icon as IconData?)
                          : (option.icon as Widget?),
                      subtitle: textOrNull(option.subtitle
                          ?.first), // Probably not idea to pick the first of the array.  not sure why the subtile is a list and not string.
                    ),
                    onPressed: () => Navigator.pop(context, option),
                  );
                },
              ),
            ],
            cancelButton: PlatformBottomSheetAction(
              isDefaultAction: true,
              child: Text(intl!.cancelLabel),
              ios: (_, target) =>
                  CupertinoBottomSheetActionData(isDefaultAction: true),
              onPressed: () => Navigator.pop(context),
            ),
          ));
}

class OptionsService implements IOptionsService {
  Map<String, IKeyedOptionsHandler> _optionsHandlers =
      <String, IKeyedOptionsHandler>{};

  void register(IKeyedOptionsHandler handler) =>
      _optionsHandlers[handler.key] = handler;

  OptionsService operator +(KeyedOptionsHandler handler) {
    register(handler);
    return this;
  }

  OptionsHandler<T?> getHandler<T>(Object key,
      {required IMetaFormFieldContext metaContext}) {
    final _key = key.baseCode;
    return (_optionsHandlers[_key!] as OptionsHandler<T?>?)
            ?.withMetaContext(metaContext) ??
        noopHandler(_key);
  }

  KeyedOptionsHandler<K, T> getKeyedHandler<K, T>(Object? key,
      {required IMetaFormFieldContext metaContext}) {
    final _key = key.baseCode;
    return (_optionsHandlers[_key!] as KeyedOptionsHandler<K, T>?)
            ?.withMetaContext(metaContext) ??
        noopKeyHandler(_key);
  }

  @override
  bool contains<K, T>(Object key) {
    final _key = key.baseCode;
    return _optionsHandlers.containsKey(_key);
  }
}

OptionsHandler<T?> noopHandler<T>(String? key) {
  _log.severe(
      "Missing/invalid handler for $T and key $key -- using noop handler");
  return OptionsHandler<T?>.ofList(
    "noop",
    canShowAll: false,
    placeholder: "Invalid configuration.",
    toOption: (T? from) => Option<T?>(null, label: "N/A"),
  );
}

KeyedOptionsHandler<K, T> noopKeyHandler<K, T>(String? key) {
  _log.severe(
      "Missing/invalid handler for $T and key $key -- using noop handler");
  return _KeyedOptionsHandler<K, T>(
    "noop",
    canShowAll: false,
    placeholder: "Invalid configuration.",
    toOption: (T? from) => from == null
        ? null
        : Option.keyed<K, T>(null, null, label: "Invalid configuration: $from"),
    toKey: (T from) => null,
  );
}

const kDefaultNumOptions = 10;

enum OptionMode { suggested, selected }

/// A base interface for helping to populate drop-downs and typeaheads.  There are a couple of common mixins to cut down
/// on boilerplate:
///
/// [InMemorySearchOptionsMixin] If you have a list of static options and then want to apply in-memory filtering, use the
/// [AlwaysFilterSearchOptionsMixin] If you always filter results, and want to force your [listAllOptions] to just execute a
/// blank search
abstract class KeyedOptionsHandler<K, T> implements IKeyedOptionsHandler<K, T> {
  final String key;

  KeyedOptionsHandler(Object? key) : key = key.baseCode!;

  const KeyedOptionsHandler.ofString(this.key);

  /// Whether or not its reasonable to display all options (or at least the first few), without any query input from
  /// the user.
  bool get canShowAll;

  /// A list of all options.
  ///
  /// Another common use case is to implement [listAllOptions] and then use the [InMemorySearchOptionsMixin] to apply full
  /// text indexing on the options.  In that case, this function should be overridden
  FutureOr<List<T>> listAllOptions([int? limit]);

  /// Filters the list of all options based on a query - the list of current selections is also provided.
  FutureOr<List<T>> filter(String? query, List currentSelections, int limit);

  KeyedAdhocBuilder<K, T>? get adhocOptionBuilder;

  /// Casts a list of inputs as this type
  List<K> castList(List input) => input.cast<K>();

  /// Converts a value [T] to a [KeyedOption]
  KeyedOption<K, T>? toOption(T? value);

  /// Produces a key, given a value
  K? toKey(T value);

  /// Loads a value given a key - because some options can be suggestive type-aheads, consumers should use
  /// [StringOptionsHandlerExtension.resolveOption] and [StringOptionsHandlerExtension.resolveAllOptions]
  FutureOr<T?> loadValue(K key);

  /// Creates a copy of this handler, for hte provided context.  Sometimes the behavior of options will change
  /// based on the surrounding context.
  KeyedOptionsHandler<K, T> withMetaContext(IMetaFormFieldContext context) =>
      this;
}

/// A base class for helping to populate drop-downs and typeaheads.
///
/// This class works with models and tries to remain free of view concerns, but the [Option] class does track title,
/// subtitle, and selection.
abstract class OptionsHandler<T> extends KeyedOptionsHandler<T, T> {
  OptionsHandler(dynamic key) : super(key);

  static OptionsHandler<T> ofEnum<T extends MLiteral<String>>(String key,
      {String? placeholder}) {
    final values = mEnumRegistry.values<T>(key);
    return OptionsHandler<T>.ofList(
      key,
      canShowAll: true,
      placeholder: placeholder,
      toOption: (T? key) => Option.ofValue<T>(key),
      allOptions: [...values],
    );
  }

  static OptionsHandler<String> ofProperty(
    IMSchemaProperty prop, {
    String placeholder = "Enter an option",
    List<String>? allOptions,
    required bool allowAdHoc,
  }) {
    if (prop is IStringSelectProperty) {
      allOptions ??= prop.options;
    }
    allOptions ??= [];
    return _OptionsHandler<String>(prop.baseCode,
        allOptions: allOptions,
        placeholder: placeholder,
        canShowAll: true,
        adhocOptionsBuilder: allowAdHoc == true
            ? ((item, {context}) => AdhocOption.ofString("$item"))
            : null,
        toOption: ((String? from) => Option.ofValue(from)));
  }

  factory OptionsHandler.ofList(
    key, {
    String? placeholder,
    Icon? inputIcon,
    required bool canShowAll,
    List<T>? allOptions,
    Mapping<T?, Option<T>?>? toOption,
    AdhocBuilder<T>? adhocOptionsBuilder,
  }) {
    return _OptionsHandler(
      key,
      placeholder: placeholder,
      inputIcon: inputIcon,
      adhocOptionsBuilder: adhocOptionsBuilder,
      canShowAll: canShowAll,
      allOptions: allOptions,
      toOption: toOption ?? ((T? value) => Option.ofValue(value)),
    );
  }

  static KeyedOptionsHandler<K, V> ofMap<K, V>(
    key, {
    String? placeholder,
    Icon? inputIcon,
    required bool canShowAll,
    bool? showImmediateOptions,
    required Map<K, V> allOptions,
    required Mapping<V?, KeyedOption<K, V>?> toOption,
    required Mapping<V, K?> toKey,
    KeyedAdhocBuilder<K, V>? adhocOptionBuilder,
  }) {
    return _KeyedOptionsHandler<K, V>(
      key,
      adhocOptionBuilder: adhocOptionBuilder,
      showImmediateOptions: showImmediateOptions ?? canShowAll,
      placeholder: placeholder,
      inputIcon: inputIcon,
      canShowAll: canShowAll,
      allOptionsMap: {...allOptions},
      toKey: toKey,
      toOption: (V? value) => toOption(value),
    );
  }

  @override
  Option<T>? toOption(T? value);

  T? loadValue(T key) => key;

  /// Creates a copy of this handler, for hte provided context.  Sometimes the behavior of options will change
  /// based on the surrounding context.
  OptionsHandler<T> withMetaContext(IMetaFormFieldContext context) => this;
}

/// A base class for helping to populate drop-downs and typeaheads
class _KeyedOptionsHandler<K, T> extends KeyedOptionsHandler<K, T>
    with InMemorySearchOptionsMixin<K, T> {
  _KeyedOptionsHandler(
    key, {
    this.placeholder,
    this.inputIcon,
    required this.canShowAll,
    this.allOptionsMap = const {},
    required Mapping<T, K?> toKey,
    required Mapping<T?, KeyedOption<K, T>?> toOption,
    this.adhocOptionBuilder,
    bool? showImmediateOptions,
  })  : _toOption = toOption,
        _toKey = toKey,
        _showImmediateOptions = showImmediateOptions,
        super(key);

  final String? placeholder;

  /// The icon for the entire input
  final Icon? inputIcon;

  /// Whether this handler allows the user to type in an ad-hoc option, or if the items in the list are restricted
  /// to only pre-defined values.
  bool get allowAdHoc => adhocOptionBuilder != null;

  /// Whether or not its reasonable to display all options
  final bool canShowAll;

  /// A list of all options
  final FutureOr<Map<K, T>> allOptionsMap;

  final bool? _showImmediateOptions;

  @override
  FutureOr<List<T>> listAllOptions([int? limit]) {
    return allOptionsMap.thenOr((m) => m.values.truncate(limit));
  }

  /// Whether to show options immediately, vs waiting for the user to type a query
  bool get showImmediateOptions => _showImmediateOptions ?? canShowAll;

  final Mapping<T?, KeyedOption<K, T>?> _toOption;
  final KeyedAdhocBuilder<K, T>? adhocOptionBuilder;

  KeyedOption<K, T>? toOption(T? value) => _toOption(value);

  @override
  FutureOr<T> loadValue(K key) =>
      allOptionsMap.thenOr(((map) => map[key]!) as T Function(Map<K, T>));

  final Mapping<T, K?> _toKey;

  @override
  K? toKey(T value) => _toKey(value);

  /// Creates a copy of this handler, for hte provided context.  Sometimes the behavior of options will change
  /// based on the surrounding context.
  KeyedOptionsHandler<K, T> withMetaContext(IMetaFormFieldContext context) =>
      this;

  @override
  IKeyedOptionsHandler<KK, VV> cast<KK, VV>() {
    return copyWith<KK, VV>();
  }

  _KeyedOptionsHandler<KK, TT> copyWith<KK, TT>({
    String? placeholder,
    Icon? inputIcon,
    bool? canShowAll,
    FutureOr<Map<KK, TT>>? allOptionsMap,
    bool? showImmediateOptions,
    Mapping<TT?, KeyedOption<KK, TT>?>? toOption,
    KeyedAdhocBuilder<KK, TT>? adhocOptionBuilder,
    Mapping<TT, KK>? toKey,
  }) {
    return _KeyedOptionsHandler<KK, TT>(
      key,
      placeholder: placeholder ?? this.placeholder,
      inputIcon: inputIcon ?? this.inputIcon,
      canShowAll: canShowAll ?? this.canShowAll,
      allOptionsMap: allOptionsMap ??
          this.allOptionsMap.thenOr((resolved) => resolved.cast<KK, TT>()),
      showImmediateOptions: showImmediateOptions ?? this._showImmediateOptions,
      toOption: toOption ??
          (this._toOption.let((fn) {
            if (fn is Mapping<TT?, KeyedOption<KK, TT>?>) {
              return fn as Mapping<TT?, KeyedOption<KK, TT>?>;
            } else {
              return ((TT? input) {
                final rs = fn(input as T?) as KeyedOption<KK, TT>?;
                return rs;
              });
            }
          }))!,
      adhocOptionBuilder: adhocOptionBuilder ??
          this.adhocOptionBuilder?.let((fn) {
            return (fn is KeyedAdhocBuilder<KK, TT>)
                ? fn as KeyedAdhocBuilder<KK, TT>
                : (String str) {
                    return fn(str) as KeyedAdhocOption<KK, TT>;
                  };
          }),
      toKey: toKey ??
          this._toKey.let((fn) {
            return (fn is Mapping<TT, KK>)
                ? fn as Mapping<TT, KK>
                : (TT key) {
                    return fn(key as T) as KK;
                  };
          })!,
    );
  }
}

/// A base class for helping to populate drop-downs and typeaheads
class _OptionsHandler<T> extends _KeyedOptionsHandler<T, T>
    implements OptionsHandler<T> {
  final Mapping<T?, Option<T>?> _toUnkeyedOption;

  _OptionsHandler(
    key, {
    String? placeholder,
    Icon? inputIcon,
    required bool canShowAll,
    List<T>? allOptions,
    AdhocBuilder<T>? adhocOptionsBuilder,
    required Mapping<T?, Option<T>?> toOption,
  })   : _toUnkeyedOption = toOption,
        super(key,
            placeholder: placeholder,
            inputIcon: inputIcon,
            adhocOptionBuilder: adhocOptionsBuilder,
            canShowAll: canShowAll,
            toKey: (_) => _,
            allOptionsMap: allOptions.keyed((_) => _),
            toOption: toOption);

  @override
  T? toKey(T value) => value;

  @override
  Option<T>? toOption(T? value) => _toUnkeyedOption(value);

  @override
  T loadValue(T value) {
    return value;
  }

  /// Creates a copy of this handler, for the provided context.  Sometimes the behavior of options will change
  /// based on the surrounding context.
  OptionsHandler<T> withMetaContext(IMetaFormFieldContext context) => this;
}

class OptionHandlerPair<K, T> {
  final KeyedOptionsHandler<K, T> baseHandler;
  final TypeaheadHandler<K, T> viewHandler;

  OptionHandlerPair(this.baseHandler, this.viewHandler);
}

extension BuildContextOptionService on BuildContext {
  OptionsService get optionsService {
    return sunny.get(context: this);
  }
}

extension IKeyedOptionHandlerCast<K, V> on IKeyedOptionsHandler<K, V> {
  KeyedOptionsHandler<K, V> cast() => this as KeyedOptionsHandler<K, V>;
}
