/*
 * flutter_platform_widgets
 * Copyright (c) 2018 Lance Johnstone. All rights reserved.
 * See LICENSE for distribution and usage details.
 */

import 'dart:ui';

import 'package:flutter/cupertino.dart';
import 'package:flutter/material.dart';
import 'package:flutter/services.dart';
import 'package:flutter/widgets.dart';
import 'package:info_x/info_x.dart';

/// Returns `child` wrapped with background and a bottom border if background color
/// is opaque. Otherwise, also blur with [BackdropFilter].
///
/// When `updateSystemUiOverlay` is true, the nav bar will update the OS
/// status bar's color theme based on the background color of the nav bar.
Widget? wrapWithBackground({
  Border? border,
  Color? backgroundColor,
  Widget? child,
  bool updateSystemUiOverlay = false,
}) {
  if (!infoX.isIOS) return child;
  Widget? result = child;
  if (updateSystemUiOverlay) {
    final bool darkBackground = backgroundColor!.computeLuminance() < 0.179;
    final SystemUiOverlayStyle overlayStyle =
        darkBackground ? SystemUiOverlayStyle.light : SystemUiOverlayStyle.dark;
    result = AnnotatedRegion<SystemUiOverlayStyle>(
      value: overlayStyle,
      sized: true,
      child: result!,
    );
  }
  final DecoratedBox childWithBackground = DecoratedBox(
    decoration: BoxDecoration(
      border: border,
      color: backgroundColor,
    ),
    child: result,
  );

  if (backgroundColor!.alpha == 0xFF) return childWithBackground;

  return ClipRect(
    child: BackdropFilter(
      filter: ImageFilter.blur(sigmaX: 10.0, sigmaY: 10.0),
      child: childWithBackground,
    ),
  );
}

PreferredSizeWidget? wrapSizedWidgetWithBackground({
  Border? border,
  Color? backgroundColor,
  PreferredSizeWidget? child,
  bool updateSystemUiOverlay = false,
}) {
  if (!infoX.isIOS) return child;
  Widget? result = child;
  if (updateSystemUiOverlay) {
    final bool darkBackground = backgroundColor!.computeLuminance() < 0.179;
    final SystemUiOverlayStyle overlayStyle =
        darkBackground ? SystemUiOverlayStyle.light : SystemUiOverlayStyle.dark;
    result = AnnotatedRegion<SystemUiOverlayStyle>(
      value: overlayStyle,
      sized: true,
      child: result!,
    );
  }
  final DecoratedBox childWithBackground = DecoratedBox(
    decoration: BoxDecoration(
      border: border,
      color: backgroundColor,
    ),
    child: result,
  );

  if (backgroundColor!.alpha == 0xFF) {
    return PreferredSize(
      preferredSize: child!.preferredSize,
      child: childWithBackground,
    );
  }

  return PreferredSize(
    preferredSize: child!.preferredSize,
    child: ClipRect(
      child: BackdropFilter(
        filter: ImageFilter.blur(sigmaX: 10.0, sigmaY: 10.0),
        child: childWithBackground,
      ),
    ),
  );
}
