import 'package:flutter/cupertino.dart';
import 'package:flutter/material.dart';
import 'package:flutter/services.dart';
import 'package:sunny_dart/json.dart';
import 'package:sunny_dart/sunny_dart.dart';
import 'package:sunny_dart/typedefs.dart';
import 'package:sunny_essentials/state.dart';
import 'package:sunny_essentials/sunny_essentials.dart';
import 'package:sunny_essentials/theme.dart';
import 'package:sunny_forms/shared/theme_helpers.dart';
import 'package:sunny_forms/sunny_form_controller.dart';
import 'package:sunny_forms/sunny_form_field.dart';
import 'package:sunny_forms/sunny_form_functions.dart';
import 'package:sunny_sdk_core/sunny_sdk_core.dart';

class TextControl<T> extends StatefulWidget {
  final JsonPath<T> attribute;
  final bool enabled;
  final bool obscureText;
  final int? minLength;
  final bool expands;
  final int? minLines;
  final int? maxLines;
  final FocusNode? focusNode;
  final ValueChanged<T>? onSubmitted;
  final bool? isRequired;
  final bool autovalidate;
  final List<ValidatorFn<T>> validators;
  final Converter<T?>? converter;
  final OnValidationError? onError;
  final dynamic placeholder;
  final bool autofocus;
  final bool? autocorrect;
  final TextEditingController? controller;
  final FormValueChanged<T>? onChange;
  final TextInputType? keyboardType;
  final TextInputAction? textInputAction;
  final T? initialValue;
  final Formatter<T>? formatter;
  final Widget? prefix;
  final Widget? suffix;
  final TextStyle? placeholderStyle;
  final TextCapitalization? textCapitalization;
  final SunnyFormFieldState<T>? state;

  TextControl.formField(
      {Key? key,
      required this.attribute,
      this.obscureText = false,
      this.keyboardType,
      this.minLength,
      this.onError,
      this.validators = const [],
      this.isRequired,
      this.autovalidate = false,
      this.initialValue,
      this.onChange,
      this.focusNode,
      this.onSubmitted,
      this.autocorrect,
      this.textInputAction,
      this.placeholder,
      this.textCapitalization,
      this.formatter,
      this.converter,
      this.controller,
      this.autofocus = false,
      this.placeholderStyle,
      this.expands = false,
      this.minLines,
      this.maxLines,
      this.enabled = true,
      this.prefix,
      this.suffix})
      : state = null,
        super(key: key ?? Key("$attribute-input"));

  TextControl.ofFormField(
    SunnyFormFieldState<T> this.state, {
    Key? key,
    this.obscureText = false,
    OnSubmit<T>? onSubmit,
    this.keyboardType,
    this.formatter,
    this.minLength,
    this.validators = const [],
    this.isRequired,
    this.autovalidate = false,
    this.controller,
    this.autocorrect,
    bool? enabled,
    this.textInputAction,
    this.placeholder,
    Converter<T?>? converter,
    this.autofocus = false,
    this.textCapitalization,
    this.placeholderStyle,
    this.expands = false,
    this.minLines,
    this.maxLines,
    this.prefix,
    this.suffix,
  })  : enabled = enabled ?? (state.isEnabled != false),
        onChange = ((value, source) {
          state.updateValue(value, source);
        }),
        focusNode = state.focusNode,
        onSubmitted = ((converted) async {
          state.updateValue(converted, AttributeSource.control);
          await onSubmit?.call(converted);
        }),
        onError = ((ValidationError error) {
          state.error = error;
        }),
        converter = converter ?? state.convert,
        attribute = state.attribute.cast(),
        initialValue = state.value,
        super(key: key ?? Key("${state.attribute}-input"));

  @override
  _TextControlState createState() => _TextControlState<T>();
}

class _TextControlState<T> extends BaseState<TextControl<T?>> {
  TextEditingController? _controller;

  @override
  void initState() {
    super.initState();
    if (widget.focusNode != null && widget.autofocus == true) {
      //Future.microtask(() => widget.focusNode.requestFocus());
    }
    final startValue = widget.formatter == null
        ? widget.initialValue?.toString()
        : widget.formatter!(widget.initialValue);
    _controller = widget.controller ??
        TextEditingController.fromValue(
          TextEditingValue(
            text: startValue ?? "",
            selection: TextSelection.collapsed(offset: startValue?.length ?? 0),
          ),
        );

    if (widget.state != null) {
      final onChange = (T? value) {
        final formatted = widget.formatter == null
            ? value?.toString()
            : widget.formatter!(value);
        _controller!.text = formatted ?? "";
      };
      stream<T?>(widget.state!.fieldValueStream, onChange);
      onChange(widget.initialValue);
    }
  }

  @override
  void dispose() {
    if (widget.controller == null) {
      _controller!.dispose();
    }
    super.dispose();
  }

  @override
  Widget build(BuildContext context) {
    final decoration = applyTheme(context,
            decoration: InputDecoration.collapsed(
              hintText: widget.placeholder?.toString(),
              floatingLabelBehavior: FloatingLabelBehavior.never,
              enabled: widget.enabled,
            ))
        .copyWith(
            contentPadding: inputPadding(
                prefix: widget.prefix != null, suffix: widget.suffix != null));

    final capitalization =
        widget.textCapitalization ?? TextCapitalization.sentences;

    
    final tf = TextField(
      key: Key("text-field-${widget.attribute}"),
      autofocus: this.widget.autofocus,
      autocorrect: this.widget.autocorrect ?? false,
      controller: _controller,
      readOnly: widget.enabled == false,
      expands: this.widget.expands,
      minLines: this.widget.minLines,
      maxLines: this.widget.maxLines,
      textCapitalization: capitalization,
      enableInteractiveSelection: true,
      enableSuggestions: true,
      toolbarOptions:
          ToolbarOptions(copy: true, cut: true, paste: true, selectAll: true),
      enabled: this.widget.enabled,
      textInputAction: this.widget.textInputAction,
      onSubmitted: (string) {
        final T? converted = convert(string);
        widget.onSubmitted?.call(converted);
        widget.focusNode
            ?.unfocus(disposition: UnfocusDisposition.previouslyFocusedChild);
      },
      style: sunnyText.input0.copyWith(color: sunnyText.input0.color?.withBrightness(context.brightness)),
      onEditingComplete: () {},
      focusNode: this.widget.focusNode,
      keyboardAppearance: context.brightness,
      keyboardType: this.widget.keyboardType ?? TextInputType.text,
      obscureText: this.widget.obscureText,
      decoration: decoration,
      onChanged: (String value) {
        try {
          final T? converted = convert(value);
          widget.onChange?.call(converted, AttributeSource.control);
        } catch (e) {
          log.fine("Error converting: $e");
        }
      },
    );

    if (widget.prefix != null || widget.suffix != null) {
      return inputWithIcon(
          input: tf, prefix: widget.prefix, suffix: widget.suffix);
    } else {
      return tf;
    }
  }

  /// Converts the value to the correct type, passing along any errors
  T? convert(String? value) {
    try {
      if (value == null) return value as T?;
      if (widget.converter != null) {
        return widget.converter!(value);
      }

      if (value is! T?) {
        illegalState<T>("Expected $T but found ${value.runtimeType}");
      }
      return value as T?;
    } catch (e) {
      if (widget.onError != null) {
        ValidationError error;
        if (e is FormatException) {
          error = ValidationError(
            path: widget.attribute,
            code: "validation.keyword.format",
            message: "$e",
            value: value,
          );
        } else {
          error = ValidationError(
            path: widget.attribute,
            code: "validation.parse",
            message: "$e",
            value: value,
          );
        }
        widget.onError!(error);
      } else {
        log.fine(e);
      }
      rethrow;
    }
  }
}

extension WidgetReIconExt on Widget {
  Widget reIcon({Color? color, double? size}) {
    final self = this;
    if (self is Icon) {
      return Icon(
        self.icon,
        color: color ?? self.color,
        size: size ?? self.size,
      );
    } else {
      return self;
    }
  }
}

Widget inputWithIcon(
    {Key? key,
    required Widget input,
    Color? color,
    Widget? prefix,
    Widget? suffix}) {
  prefix = prefix?.reIcon(color: color, size: 20);
  suffix = suffix?.reIcon(color: color, size: 20);

  if (prefix == null && suffix == null) {
    return input;
  } else {
    return Stack(
      key: key,
      fit: StackFit.passthrough,
      alignment: AlignmentDirectional.centerStart,
      children: [
        if (prefix != null) prefix,
        input,
        if (suffix != null) Positioned(right: 0, child: suffix)
      ],
    );
  }
}
