import 'dart:async';

import 'package:flexidate/flexidate.dart';
import 'package:sunny_dart/helpers/logging_mixin.dart';
import 'package:sunny_dart/sunny_dart.dart';
import 'package:sunny_sdk_core/services/sunny.dart';
import 'package:sunny_service_stubs/meta.dart';
import 'package:sunny_service_stubs/models.dart';
import 'package:timezone/timezone.dart';

import 'smart_date_fact_query.dart';
import 'smart_date_query.dart';

class SmartDateService with LoggingMixin {
  final ISmartDateApi api;
  final IFactService factService;

  SmartDateService(this.api, this.factService);

  Future<List<SmartDateQueryResult>> parseDates(String text,
      {bool isFlexiDate = false,
      IContact? contact,
      FutureOr<List<IFact>>? facts,
      Location? location}) async {
    try {
      final _facts = (await facts ?? []);
      final matches = await factService.findFactSchemas(text);

      var result =  await _resolveAndQuery(
        text: text,
        isFlexiDate: isFlexiDate,
        contact: contact,
        facts: _facts,
        matches: matches,
        location: location,
      );

      return result;
    } catch (e, stack) {
      log.severe("Error fetching date results $e", e, stack);
      rethrow;
    }
  }

  /// Typically, the user types in a query, like "week before birthday", the system extracts "Birthday", constructs a
  /// query that can be sent to the server to be resolved.  In the case of default values, we don't need to do the
  /// extra interpolation because we the source of the query is the date itself.constructed the query to be the date exactly.
  Future<List<SmartDateQueryResult>> resolveMetaDateResults(
      {required Iterable<MetaDateRef> refs,
      IContact? contact,
      FutureOr<List<IFact>>? facts,
      Location? location}) async {
    try {
      final _facts = await facts!;
      final Iterable<FactDateSchemaQuery> matches =
          (await factService.resolveMetaDates(refs)).map((_) => _.toQuery());
      final Iterable<SmartDateQueryResult> metaDateMatches = matches.expand(
        (match) {
          return _resolveFactDates(
                  metaDate: match,
                  location: location,
                  facts: _facts,
                  contact: contact)
              .map((contactFactResult) =>
                  contactFactResult.toSingleQueryResult())
              .ifEmpty(() => [
                    UnresolvedFactDateQueryResult.ofSchemaQuery(match, contact!)
                  ]);
        },
      );

      return [...metaDateMatches];
    } catch (e) {
      log.severe("Error fetching date results $e", e);
      rethrow;
    }
  }

  /// Given a FactMetaDate [metaDate], looks for one or more facts for the contact that provide a value
  /// for that date.  These results are returned as a list of [FactSmartDateQuerySource].  These can be
  /// then passed to the date parsing engine, or in the case of default values, simply resolved.
  List<FactSmartDateQuerySource> _resolveFactDates({
    FactDateSchemaQuery? metaDate,
    Location? location,
    List<IFact>? facts,
    IContact? contact,
  }) {
    // Attempt to resolve against know facts and/or dates
    final matchingFacts = [
      ...facts
          .orEmptyIter()
          .where((fact) => fact.mmeta.mtype == metaDate!.factSchema.ref)
          .mapNotNull((fact) {
        final mwf = metaDate!.withFact(fact);

        /// If the record doesn't have a valid date, this next line will return null.
        return mwf.toSmartDateQuery(contact, location!)!;
      }),
    ];

    return matchingFacts;
  }

  Future<List<SmartDateQueryResult>> _resolveAndQuery(
      {String? text,
      bool isFlexiDate = false,
      IContact? contact,
      List<IFact>? facts,
      required Iterable<FactDateSchemaQuery> matches,
      Location? location}) async {
    final unresolved = <FactDateSchemaQuery>[];
    final queries = <SmartDateQuerySource>{};

    for (final match in matches) {
      // Attempt to resolve against know facts and/or dates
      final matchingFacts = _resolveFactDates(
          metaDate: match, location: location, facts: facts, contact: contact);
      if (matchingFacts.isEmpty && contact != null) {
        unresolved.add(match);
      } else {
        queries.addAll(matchingFacts);
      }
    }

    if (queries.isEmpty && text.isNotNullOrBlank) {
      queries.add(TermSource(text));
    }

    location ??= sunnyLocalization.userLocation;

    final queryResults = await executeQueries(
      queries,
      isFlexiDate: isFlexiDate,
      existingResults: unresolved
          .map((_) => UnresolvedFactDateQueryResult.ofSchemaQuery(_, contact!)),
      location: location!,
    );
    return queryResults;
  }

  Future<List<SmartDateQueryResult>> executeQueries(
    Iterable<SmartDateQuerySource> queries, {
    bool isFlexiDate = false,
    Iterable<SmartDateQueryResult>? existingResults,
    required Location location,
  }) async {
    final List<List<SmartDateQueryResult>> queryResults = (await Future.wait(
      queries.map((query) async {
        final flexiDate = FlexiDate.tryFrom(query.smartDateQuery);
        // Attempt to parse a flexiDate first.
        if (flexiDate != null && flexiDate.isValid && isFlexiDate == true) {
          return [SmartDateQueryResult.ofFlexiDate(flexiDate)];
        }
        final result = await api.parseSmartDate(
          IParseSmartDateRequest.of(
            input: query.smartDateQuery,
            timeZone: location,
          ),
        );

        final expanded = [
          ...result.results!.expand((r) {
            final qr = query.toQueryResult(r);
            return qr;
          }),
          if (result.results!.isEmpty == true && flexiDate != null)
            SmartDateQueryResult.ofFlexiDate(flexiDate),
        ];
        return expanded;
      }),
    ));

    final allResults = [
      ...queryResults.flatten(),
      ...?existingResults,
    ];
    return allResults;
  }
}

SmartDateService get smartDateService => sunny.get();

// TZDateTime tzDateTime(DateTime dateTime) =>
//     dateTime is TZDateTime ? dateTime : TZDateTime.from(dateTime, sunnyLocalization.userLocation);

// String smartDateFormat(DateTime dateTime, {bool formatTime, bool useToday = true}) {
//   if (dateTime == null) return null;
//   formatTime ??= hasTime(dateTime);
//   TZDateTime tzToFormat = tzDateTime(dateTime);
//   if (!formatTime) {
//     final now = TZDateTime.now(tzToFormat.location);
//     final date = tzToFormat;
//     if (now.day == date.day && useToday == true) {
//       // Same day
//       return "today";
//     } else if (now.year == date.year) {
//       return dateFormatNoYear.format(date);
//     } else {
//       return dateFormatWithYear.format(date);
//     }
//   } else {
//     return smartDateTimeFormat(tzToFormat);
//   }
// }
//
// String smartDateTimeFormat(DateTime dateTime) {
//   final TZDateTime toFormat = tzDateTime(dateTime);
//   final now = TZDateTime.now(toFormat.location);
//   final date = toFormat.toLocal();
//
//   if (now.day == date.day) {
//     return formatDateTimeSameDay(date);
//   } else if (now.year == date.year) {
//     return formatDateTimeSameYear(date);
//   } else {
//     return formatDateTimeWithYear(date);
//   }
// }

// final dateFormatWithYear = DateFormat("MMM d, yyyy");
// final dateFormatNoYear = DateFormat("EEE, MMM d");
// final dateFormatMonthDay = DateFormat("MMM d");
//
// String formatDateTimeWithYear(TZDateTime dateTime) => _withLocation(_dateTimeFormatWithYear, dateTime);
//
// String formatDateTimeSameYear(TZDateTime dateTime) => _withLocation(_dateTimeFormatSameYear, dateTime);
//
// String formatDateTimeSameDay(TZDateTime dateTime) => _withLocation(_dateTimeFormatSameDay, dateTime);
//
// String _withLocation(DateFormat format, TZDateTime toFormat) {
//   // To format it properly, we need to convert it to a local time
//   toFormat.toLocal();
//   return format.format(toFormat) + " " + toFormat.location.currentTimeZone.abbr;
// }
//
// final _dateTimeFormatWithYear = DateFormat("MMM d, yyyy 'at' h:mma");
// final _dateTimeFormatSameYear = DateFormat("EEE, MMM d 'at' h:mma");
// final _dateTimeFormatSameDay = DateFormat("'today at 'h:mm a");
