import 'dart:async';

import 'package:flutter/material.dart';
import 'package:flutter/widgets.dart';
import 'package:sunny_dart/json.dart';
import 'package:sunny_dart/sunny_dart.dart' hide Converter;
import 'package:sunny_dart/typedefs.dart';
import 'package:sunny_essentials/state.dart';
import 'package:sunny_forms/keynames.dart';
import 'package:sunny_forms/meta/meta.dart';
import 'package:sunny_forms/notifier.dart';
import 'package:sunny_forms/standard_column.dart';
import 'package:sunny_forms/sunny_form_controller.dart';
import 'package:sunny_forms/sunny_form_theme.dart';
import 'package:sunny_forms/text_widgets.dart';
import 'package:sunny_forms/widget_extensions.dart';
import 'package:sunny_sdk_core/sunny_sdk_core.dart';

import 'keynames.dart';

typedef FormValueChanged<T> = void Function(T value, AttributeSource source);

typedef SunnyFormFieldValidator<T> = FutureOr<ValidationError> Function(
    T value);

/// Similar to [FormField], but extended to work with Sunny metadata and styling.
///
/// A [SunnyFormField] represents a form control that manages data at a path [attribute].  This widget renders the control,
/// label, and error widgets.  The underlying value is tracked within [SunnyFormFieldState] / but if there is a
/// [FormController] in the other scope, this field will publish any changes there.
///
///
class SunnyFormField<T> extends StatefulWidget {
  final dynamic label;

  /// Provides a hook for handling editable labels
  final Producer<String>? onLabelTap;
  final ValueChanged<String>? onLabelChange;

  final dynamic placeholder;
  final JsonPath<T> attribute;
  final bool isRequired;
  final List<SunnyFormFieldValidator>? validators;

  /// This builder should only build the raw input, without any extra stuff, like errors, help text,
  /// labels, or otherwise
  final SunnyFormFieldBuilder<T>? builder;
  final bool? isReadOnly;
  final bool? enabled;
  final AttributeAccepter<T?>? onReadOnlyTapped;
  final T? initialValue;
  final FieldConverter<T?>? converter;
  final FocusNode? focusNode;
  final InputDecoration? decoration;
  final FormValueChanged<T?>? onChange;
  final FormController? form;

  final bool captureNestedPaths;

  SunnyFormField({
    Key? key,
    String discriminator = "",
    required this.label,
    required dynamic attribute,
    this.placeholder,
    required this.isRequired,
    this.onLabelTap,
    this.onLabelChange,
    this.decoration,
    this.converter,
    required this.builder,
    bool? captureNestedPaths,
    this.enabled = true,
    this.isReadOnly = false,
    this.focusNode,
    this.onReadOnlyTapped,
    this.initialValue,
    this.onChange,
    this.form,
    this.validators,
  })  : captureNestedPaths = captureNestedPaths ?? true,
        attribute = JsonPath.of(attribute),
        super(key: key ?? Key("$discriminator-${attribute ?? 'unknown'}"));

  /// This constructor expects form metadata to be the source for the field rendering, as well
  /// as a [SunnyForm] to be present.
  factory SunnyFormField.meta(
    BuildContext context, {
    Key? key,
    // Optional
    String? label,

    // Optional
    String? fallbackLabel,

    // Optional
    String? placeholder,
    dynamic basePath,
    // If this isn't provided, it will be lookup up using a Provider
    MetaFormContext? formMetadata,
    required dynamic attribute,
    bool? captureNestedPaths,
    bool isReadOnly = false,
    bool? isRequired,
    AttributeAccepter<T?>? onReadOnlyTapped,
    FormValueChanged<T?>? onChange,
  }) {
    var _basePath = JsonPath.of(basePath);
    var _attribute = JsonPath.of(attribute);
    final fullPath = (_basePath + _attribute).cast<T>();
    final meta = formMetadata ?? context.formMetadata;
    final metaField = meta.path(_basePath, _attribute);
    return SunnyFormField<T>(
      key: key ?? Key("sunnyForm-field-${fullPath.toKey()}"),
      attribute: fullPath,
      builder: (context, state) {
        return MetaFormControl<T>.ofJsonPath(
          path: fullPath,
          metaForm: meta,
          fieldState: state,
        );
      },
      captureNestedPaths: captureNestedPaths == true,
      isReadOnly: isReadOnly,
      isRequired: isRequired ??
          metaField.isRequired ??
          metaField.formConfig()?.isRequired == true,
      onReadOnlyTapped: onReadOnlyTapped,
      onChange: onChange,
      label: metaField.resolveLabel(
          labelOverride: label, fallbackLabel: fallbackLabel),
      placeholder:
          metaField.resolvePlaceholder(placeholderOverride: placeholder),
    );
  }

  @override
  State<StatefulWidget> createState() {
    return _SunnyFormFieldState<T>();
  }
}

abstract class SunnyFormFieldState<T> {
  T? get value;

  Stream<T?> get fieldValueStream;

  T? convert(String? input);

  bool? captureNestedPaths;

  set error(ValidationError error);

  void updateValue(T? value, AttributeSource source);

  bool? isEnabled;
  List<ValidationError?>? errors;

  /// The FocusNode associated with this form element.  Managed primarily by the containing form
  FocusNode? focusNode;

  JsonPath<T> get attribute;

  FutureOr<List<ValidationError>> validate();

  FormController? get form;

  SunnyFormFieldState<R> cast<R>();
}

class DelegatingSunnyFormFieldState<T> implements SunnyFormFieldState<T> {
  final SunnyFormFieldState _delegate;

  DelegatingSunnyFormFieldState(this._delegate);

  @override
  bool? get captureNestedPaths => _delegate.captureNestedPaths;

  @override
  set captureNestedPaths(bool? processNestedPaths) =>
      _delegate.captureNestedPaths = processNestedPaths;

  @override
  List<ValidationError?>? get errors => _delegate.errors;

  @override
  FocusNode? get focusNode => _delegate.focusNode;

  @override
  bool? get isEnabled => _delegate.isEnabled;

  @override
  T? get value => _delegate.value as T?;

  @override
  JsonPath<T> get attribute => _delegate.attribute.cast();

  @override
  FormController? get form => _delegate.form;

  @override
  set errors(List<ValidationError?>? _errors) => _delegate.errors = _errors;

  @override
  set focusNode(FocusNode? _focusNode) => _delegate.focusNode = _focusNode;

  @override
  set isEnabled(bool? _isEnabled) => _delegate.isEnabled = _isEnabled;

  @override
  set error(ValidationError? error) {
    if (error != null) {
      errors = [error];
    } else {
      errors = [];
    }
  }

  @override
  void updateValue(T? value, AttributeSource source) {
    _delegate.updateValue(value, source);
  }

  @override
  SunnyFormFieldState<R> cast<R>() =>
      DelegatingSunnyFormFieldState<R>(this._delegate);

  @override
  FutureOr<List<ValidationError>> validate() => _delegate.validate();

  @override
  Stream<T?> get fieldValueStream {
    return _delegate.fieldValueStream.cast();
  }

  @override
  T? convert(String? input) => _delegate.convert(input) as T?;
}

class _SunnyFormFieldState<T> extends BaseState<SunnyFormField<T>>
    with TickerProviderStateMixin
    implements SunnyFormFieldState<T> {
  FocusNode? _focusNode;

  // ignore: unused_field
  bool _isFocusOwner = false;

  @override
  FocusNode? get focusNode {
    if (_focusNode != null) return _focusNode!;
    if (widget.focusNode != null) {
      form?.addFocusNode(attribute, widget.focusNode);
      _focusNode = widget.focusNode;
    } else {
      _focusNode = form?.getFocusNode(attribute) ?? FocusNode();
    }

    if (_focusNode == null) {
      _isFocusOwner = true;
      _focusNode = FocusNode();
    }

    return _focusNode!;
  }

  @override
  set focusNode(FocusNode? newFocusNode) {
    assert(newFocusNode != null, "Provided a null FocusNode for $attribute");
    assert(widget.focusNode == null || widget.focusNode == newFocusNode,
        "Attempting to override a focus node for $attribute that the caller specified");
    assert(_focusNode == null || _focusNode == newFocusNode,
        "Already provided a focusNode for $attribute and attempting to update it");
    form?.addFocusNode(attribute, newFocusNode);
    _focusNode = newFocusNode;
  }

  List<ValidationError?>? _errors;

  @override
  bool? isEnabled;

  /// The internal form control will set this at render time. An example would be a range control that has two
  /// internal properties that need to be listened to and rendered as a single error.
  ///
  /// Probably a better way of doing this.
  @override
  bool? captureNestedPaths;

  T? _value;

  @override
  T? get value => _value;

  set value(T? _value) {
    updateValue(_value, AttributeSource.control);
  }

  @override
  void updateValue(T? _value, AttributeSource source) {
    this._value = value;

    /// Not sure this should really go here
    widget.onChange?.call(_value, source);

    /// If there's a form, bind it
    context.formOrNull(listen: false)?.set(
          attribute,
          _value,
          source: source,
        );
  }

  @override
  List<ValidationError?>? get errors => _errors;

  @override
  set errors(List<ValidationError?>? errors) {
    this._errors = errors;

    /// If there's a form, bind it
    context.formOrNull(listen: false)?.let((form) {
      if (errors.isNullOrEmpty) {
        form.clearError(attribute, source: AttributeSource.control);
      } else {
        errors!.forEach((err) {
          form.addError(attribute, err, source: AttributeSource.control);
        });
      }
    });
  }

  /// Updates the errors without triggering updates
  void setErrorQuiet(List<ValidationError?> errors) {
    this._errors = errors;
  }

  @override
  set error(ValidationError? error) {
    this.errors = [if (error != null) error];
  }

  @override
  void initState() {
    super.initState();

    captureNestedPaths = widget.captureNestedPaths;
    form = widget.form ?? context.formOrNull(listen: false);

    if (widget.focusNode != null) {
      focusNode = widget.focusNode!;
      form?.addFocusNode(attribute, focusNode);
    } else {
      form?.getFocusNode(attribute)?.let((node) {
        focusNode = node;
      });
    }
    focusNode!.addListener(_onFocusChanged);
    if (form != null) {
      var subscription = form!.stream.where((attr) {
        // Ignore all updates that initiated from the control itself.
        return attr.isError ||
            attr.source != AttributeSource.control ||
            attr.clearErrors == true;
      }).listen((attr) {
        final shouldProcess =
            (captureNestedPaths! && attr.path!.startsWith(attribute)) ||
                (attr.path == attribute);
        if (shouldProcess) {
          setState(() {
            if (attr.clearErrors) {
              setErrorQuiet([]);
            } else if (attr.source == AttributeSource.control) {
              setErrorQuiet(
                  [if (attr.error != null) attr.error as ValidationError]);
            } else {
              error = attr.error as ValidationError?;
            }
            _value = attr.value as T?;
          });
        } else {
          // log.info("Ignoring ${attr.path} cuz it aint us ($attribute)");
        }
      }, cancelOnError: false);

      disposers.add(subscription.cancel);
    }

    _value = widget.initialValue ?? form?.get(attribute);
    error = form == null ? null : form?.errors[attribute];
  }

  @override
  void dispose() {
    focusNode?.removeListener(_onFocusChanged);
//    if (_isFocusOwner) {
//      _focusNode.dispose();
//    }
    super.dispose();
  }

  @override
  Future<List<ValidationError>> validate() async {
    final errors = await Future.wait([
      for (var validator in widget.validators.orEmpty())
        Future.value(validator(value)),
    ]);
    return [
      if (widget.isRequired && value == null)
        ValidationError.ofString(attribute, "This field is required"),
      ...errors,
    ];
  }

  @override
  FormController? form;

  @override
  JsonPath<T> get attribute => widget.attribute;

  /// Contains updates to the value that come from within the form itself, as opposed to being modified within the
  /// control itself.
  @override
  Stream<T?> get fieldValueStream {
    form = widget.form ?? context.formOrNull(listen: false);
    if (form != null) {
      return form!.stream.where((attr) {
        // Ignore all updates that initiated from the control itself.
        final shouldProcess =
            (captureNestedPaths == true && attr.path!.startsWith(attribute)) ||
                (attr.path == attribute);
        return attr.source == AttributeSource.model &&
            attr.error == null &&
            shouldProcess;
      }).map((attr) => attr.value as T?);
    } else {
      return Stream.empty();
    }
  }

  @override
  Widget build(BuildContext context) {
    final form = SunnyForm.of(context, listen: true);
    return _buildFormControl(context, form);
  }

  bool get _isFocused => focusNode!.hasFocus;

  Widget _buildFormControl(BuildContext context, FormController? form) {
    if (widget.isReadOnly != true) {
      return column([
        if (widget.label != null)
          SunnyFieldLabel(
            isRequired: widget.isRequired,
            path: widget.attribute,
            label: _widgetLabel,
            onTap: widget.onLabelTap,
            onChange: widget.onLabelChange,
            isFocused: _isFocused,
          ),
        SunnyFieldContainer(
          child: widget.builder!(context, this),
          path: widget.attribute,
        ),
        SunnyFieldError(
            path: widget.attribute, errors: this.errors, vsync: this),
      ]);
    }
    return column(
      [
        if (widget.label != null) TertiaryText(_widgetLabel),
        if (widget.isReadOnly == true && widget.onReadOnlyTapped != null)
          GestureDetector(
            onTap: () =>
                widget.onReadOnlyTapped!(widget.attribute, widget.initialValue),
            child: PrimaryText(widget.initialValue?.toString() ?? ""),
          ),
        if (widget.isReadOnly == true && widget.onReadOnlyTapped == null)
          PrimaryText(widget.initialValue?.toString() ?? ""),
      ],
    );
  }

  String get _widgetLabel {
    final label = widget.label;
    if (label == null) {
      return "";
    } else if (label is String) {
      return label;
    } else {
      return illegalState("Bad type: ${label.runtimeType}");
    }
  }

//  _resolveLabel(SunnyIntl intl, SunnyForm form) {
//    var label = widget.label;
//    if (label is Function) {
//      label = label?.call(intl);
//    }
//    return label ?? form.resolveLabel(intl, widget.attribute);
//  }
//
//  _resolvePlaceholder(SunnyIntl intl, SunnyForm form) {
//    var placeholder = widget.placeholder;
//    if (placeholder is Function) {
//      placeholder = placeholder?.call(intl);
//    }
//    return placeholder ?? form.resolvePlaceholder(intl, widget.attribute);
//  }

  void _onFocusChanged() {
    setState(() {});
  }

  @override
  // ignore: use_to_and_as_if_applicable
  SunnyFormFieldState<X> cast<X>() {
    return DelegatingSunnyFormFieldState<X>(this);
  }

  @override
  T? convert(String? input) => widget.converter?.call(input) ?? input as T;
}

typedef SunnyFormFieldBuilder<T> = Widget Function(
    BuildContext context, SunnyFormFieldState<T> state);

class SunnyFieldLabel extends StatefulWidget {
  final bool? isFocused;
  final String label;
  final Producer<String>? onTap;
  final ValueChanged<String>? onChange;
  final bool isRequired;

  SunnyFieldLabel(
      {required JsonPath path,
      this.isFocused,
      required this.isRequired,
      required this.label,
      this.onTap,
      this.onChange})
      : super(key: Key(sunnyFieldLabelKey(path)));

  @override
  _SunnyFieldLabelState createState() => _SunnyFieldLabelState();
}

class _SunnyFieldLabelState extends State<SunnyFieldLabel> {
  String? _label;

  @override
  void initState() {
    super.initState();
    _label = widget.label;
  }

  @override
  Widget build(BuildContext context) {
    final theme = Theme.of(context);
    final changeLabel = widget.onTap == null
        ? null
        : () async {
            final newLabel = await widget.onTap!();
            setState(() {
              this._label = newLabel;
            });
            widget.onChange?.call(newLabel);
          };

    final baseStyle = theme.inputDecorationTheme.labelStyle ?? TextStyle();
    return AnimatedDefaultTextStyle(
      style: baseStyle.copyWith(
        color: widget.isFocused == true
            ? theme.inputDecorationTheme.focusColor
            : null,
      ),
      child: row([
        Text(_label ?? "").tappable(changeLabel),
        if (widget.isRequired &&
            SunnyFormTheme.of(context).starForRequiredFields)
          Text(" *")
      ]),
      duration: 200.ms,
    );
  }
}

class SunnyFieldError extends StatelessWidget {
  final bool? isFocused;
  final List<ValidationError?>? errors;
  final JsonPath path;
  final bool hasError;
  final TickerProvider? vsync;

  SunnyFieldError(
      {required this.path,
      this.isFocused,
      required this.errors,
      required this.vsync})
      : hasError = errors.isNotNullOrEmpty,
        super(key: Key(sunnyFieldErrorKey(path)));

  SunnyFieldError.simple({required JsonPath path, String? error})
      : this(
          path: path,
          errors: [ValidationError.ofString(path, error)],
          vsync: null,
          isFocused: true,
        );

  @override
  Widget build(BuildContext context) {
    Widget _error() {
      return hasError
          ? Container(
              padding: EdgeInsets.only(top: 10),
              alignment: Alignment.centerLeft,
              child: column([
                for (final e in errors!) ErrorText.ofValidationError(e!),
              ]),
            )
          : emptyBox;
    }

    return vsync == null
        ? _error()
        : AnimatedSize(
            key: Key("animated-error-${path.toKey()}"),
            child: _error(),
            duration: 300.ms,
            vsync: vsync!,
          );
  }
}

class SunnyFieldContainer extends StatelessWidget {
  final Widget child;

  SunnyFieldContainer({required JsonPath path, required this.child})
      : super(key: Key(sunnyFieldContainerKey(path)));

  @override
  Widget build(BuildContext context) {
    return child;
  }
}
