import 'package:flutter/cupertino.dart';
import 'package:flutter/material.dart';
import 'package:sunny_dart/sunny_dart.dart';
import 'package:sunny_sdk_core/mverse/m_model.dart';
import 'package:mobx/mobx.dart' hide Listenable;
import 'package:observable_collections/observable_collections.dart';
import 'package:sunny_forms/sunny_list/sunny_animated_list.dart';

/// Keeps a Dart List in sync with an AnimatedList.
///
/// The [insert] and [removeAt] methods apply to both the internal list and the
/// animated list that belongs to [listKey].
///
/// This class only exposes as much of the Dart List API as is needed by the
/// sample app. More list methods are easily added, however methods that mutate the
/// list must make the same changes to the animated list in terms of
/// [AnimatedListState.insertItem] and [AnimatedList.removeItem].
class ListModel<E> extends ChangeNotifier with LoggingMixin {
  ListModel.ofListKey({
    this.removedItemBuilder,
    this.container,
    required GlobalKey<AnimatedListState> this.listKey,
    this.attribute,
    this.animationType = AnimationType.slide,
    Iterable<E>? initialItems,
  }) : _items = [...?initialItems] {
    this.count.update(_items.length);
  }

  /// Creates a list model that represents keys in a map
  ListModel.ofObservableMap({
    required ObservableMap<E, dynamic> map,
    this.container,
    this.listKey,
    this.animationType = AnimationType.slide,
    this.attribute,
  }) : _items = map.keys.toList() {
    // Sync changes with the observable map keys
    _dispose = map.observe((change) {
      updateAll(map.keys);
    });
  }

  final GlobalKey<AnimatedListState>? listKey;

  AnimatedListState? get _animatedList => listKey!.currentState;

  RemovedItemBuilder<E>? removedItemBuilder;

  final MModel? container;
  final String? attribute;
  final List<E> _items;
  final AnimationType animationType;
  VoidCallback? _dispose;
  final SingleValue<int> count = SingleValue<int>(0);

  Iterable<E> get items => List.unmodifiable(_items).cast<E>();

  bool get isEmpty => length == null || length == 0;

  bool get isNotEmpty => !isEmpty;

  void dispose() {
    super.dispose();
    _dispose?.call();
  }

  void replace(int index, E item) {
    removeAt(index);
    insert(index, item);
  }

  void insert(int index, E item) {
    runInAction(() {
      _items.insert(index, item);
      // This is required because mmodel instances don't automatically track changes
      // made directly to collections
      if (container != null && attribute != null) {
        container![attribute!] = _items.cast<E>();
      }
    });
    _animatedList?.insertItem(index);
    _sync();
  }

  void add(E item) {
    this.insert(0, item);
  }

  bool updateAll(final Iterable<E> newItems) {
    try {
      bool changed = false;

      final currSize = this._items.length;
      final newList = newItems.toList();
      var i = 0;
      for (; i < newItems.length; i++) {
        final newItem = newList[i];
        if (_items.length > i) {
          final existingItem = _items[i];
          if (existingItem != newItem) {
            changed = true;
            removeAt(i);
            insert(i, newItem);
          }
        } else {
          changed = true;
          insert(_items.length, newItem);
        }
      }

      final trimSize = currSize - i;
      if (trimSize > 0) {
        Iterable.generate(trimSize).forEach((_) => removeLast());
        changed = true;
      }

      return changed;
    } catch (e) {
      log.info("Error updating list state: $e");
      rethrow;
    }
  }

  E removeLast() {
    return removeAt(_items.length - 1);
  }

  E removeAt(int index) {
    try {
      final E removedItem = runInAction(() {
        final removed = _items.removeAt(index);
        if (container != null && attribute != null) {
          container![attribute!] = _items;
        }
        return removed;
      });
      if (removedItem != null) {
        if (_animatedList == null) {
          log.info("Something weird is going on!");
        }
        _animatedList?.removeItem(index,
            (BuildContext context, Animation<double> animation) {
          final removed =
              removedItemBuilder!(context, removedItem, index, animation);
          return removed;
        });
      }
      _sync();
      return removedItem;
    } catch (e) {
      log.info("Error updating list state: $e");
      rethrow;
    }
  }

  _sync() {
    count.update(_items.length);
    notifyListeners();
  }

  int? get length => count.value;

  E operator [](int index) => _items[index];

  int indexOf(E item) => _items.indexOf(item);

  ListModel<T> cast<T>() {
    return ListModel.ofListKey(
      removedItemBuilder: (context, T t, int index, animation) {
        return this.removedItemBuilder!(context, t as E, index, animation);
      },
      container: container,
      listKey: listKey!,
      attribute: attribute,
      initialItems: items.cast<T>(),
    );
  }
}

typedef RemovedItemBuilder<T> = Widget Function(
    BuildContext context, T item, int index, Animation<double> animation);
