import 'dart:async';
import 'dart:math' hide log;

import 'package:collection_diff/collection_diff.dart';
import 'package:collection_diff/list_diff_model.dart';
import 'package:collection_diff_worker/collection_diff_worker.dart';
import 'package:flutter/material.dart';
import 'package:flutter/widgets.dart';
import 'package:flutter_slidable/flutter_slidable.dart';
import 'package:flutter_staggered_grid_view/flutter_staggered_grid_view.dart';
import 'package:observable_collections/observable_collections.dart';
import 'package:sunny_dart/sunny_dart.dart';
import 'package:sunny_dart/typedefs.dart';
import 'package:sunny_essentials/container.dart';
import 'package:sunny_essentials/state.dart';
import 'package:sunny_forms/notifier.dart';
import 'package:sunny_forms/sunny_colors.dart';
import 'package:sunny_forms/sunny_list/record_stream_extensions.dart';
import 'package:sunny_forms/sunny_list/sunny_animated_list.dart';
import 'package:sunny_forms/sunny_list/sunny_grid.dart';
import 'package:sunny_forms/sunny_list/sunny_list.dart';
import 'package:sunny_forms/sunny_list/sunny_list_header.dart';
import 'package:sunny_forms/sunny_list/sunny_sliver_animated_list.dart';
import 'package:sunny_forms/sunny_list/sunny_sliver_list.dart';
import 'package:sunny_sdk_core/mverse.dart';

typedef KeyGenerator<T> = String? Function(T input);
typedef SunnyListWidgetBuilder<T> = Widget Function(
    BuildContext context, SunnyListBaseState<T> list);
typedef SwipeActionBuilder<T> = Widget Function(BuildContext context, T item);
typedef ListTransform<F, T> = Future<Iterable<T>> Function(Iterable<F> source);

typedef SunnyListDelegate<T> = SunnyListBaseState<T> Function();

SunnyListDelegate<T?> animatedList<T>() => () => SunnyListState<T>();

SunnyListDelegate<T?> sliverList<T>() => () => SunnySliverListState<T>();

SunnyListDelegate<T?> sliverAnimatedList<T>() =>
    () => SunnySliverAnimatedListState<T>();

typedef StaggeredTileBuilder<T> = StaggeredTile Function(T item, int index);

SunnyListDelegate<T?> grid<T>({
  /// GRID PROPERTIES
  double mainAxisSpacing = 5.0,
  double crossAxisSpacing = 5.0,
  int crossAxisCount = 12,
  double tileHeight = 45.0,
  int tileWidthUnit = 4,
}) =>
    () => SunnyGridState<T>.ofFixedExtent(
          mainAxisSpacing: mainAxisSpacing,
          crossAxisCount: crossAxisCount,
          crossAxisSpacing: crossAxisSpacing,
          tileHeight: tileHeight,
          tileWidthUnit: tileWidthUnit,
        );

SunnyListDelegate<T?> staggeredGrid<T>({
  /// GRID PROPERTIES
  required double mainAxisSpacing,
  required double crossAxisSpacing,
  required int crossAxisCount,
  required StaggeredTileBuilder<T?> staggeredBuilder,
}) =>
    () => SunnyGridState<T>(
        mainAxisSpacing: mainAxisSpacing,
        crossAxisCount: crossAxisCount,
        crossAxisSpacing: crossAxisSpacing,
        staggeredBuilder: staggeredBuilder);

/// A base class widget that dynamically responds to changes in a list or map
class SunnyList<T> extends StatefulWidget {
  final String id;

  /// The list that will be observed.
  final Getter<ValueStream<Iterable<T?>>> vstream;

  final List<T>? _start;

  final bool showDividers;

  /// The list that will be observed.
  final SunnyListDelegate<T?> state;

  /// Key generator
  final KeyGenerator<T>? keyGenerator;

  /// Key generator
  final DiffEquality diffEquality;

  /// Item builder, also handles animations
  final ListTileBuilder<T>? itemBuilder;

  final SunnyListWidgetBuilder<T>? header;

  /// A scroll controller to use with the list, if you need to access it
  final ScrollController? scrollController;

  /// Padding around the outside edge of the entire list
  final EdgeInsets? padding;

  /// A callback for whenever the list changes
  final Consumer<Iterable<T>>? onChange;

  /// The type of animation to use for new items on the list.
  final AnimationType animationType;

  /// Animation used to dismiss items.
  final AnimationType dismissAnimationType;

  /// Empty state builder
  final SunnyListWidgetBuilder<T>? emptyState;

  /// Swipe actions to attach to the starting edge
  final List<SwipeActionBuilder<T>> primarySwipeActions;

  /// Wraps each list item in a state tracker that will be send refreshes
  final bool useStateTracker;

  /// Swipe actions to attach to the trailing edge
  final List<SwipeActionBuilder<T>> secondarySwipeActions;

  /// Whether this list is the primary list
  final bool? primary;

  /// Whether this list is shrinkwrapped
  final bool? shrinkWrap;

  final bool reverse;

  final bool? refreshWithPage;

  /// Reverses the list
  SunnyList.ofVStream({
    required this.id,
    this.itemBuilder,
    this.showDividers = false,
    this.keyGenerator,
    this.diffEquality = const DiffEquality(),
    SunnyListDelegate<T?>? type,
    required this.vstream,
    this.primary,
    this.onChange,
    this.reverse = false,
    this.useStateTracker = true,
    this.scrollController,
    this.refreshWithPage,
    this.padding,
    this.animationType = AnimationType.size,
    this.dismissAnimationType = AnimationType.slide,
    this.shrinkWrap,
    this.emptyState,
    this.primarySwipeActions = const [],
    this.secondarySwipeActions = const [],
    this.header,
    List<T>? start,
  })  : _start = start,
        state = type ?? (() => SunnyListState()),
        super(key: Key(id));

  @protected
  SunnyList.ofList({
    required this.id,
    this.itemBuilder,
    SunnyListDelegate<T?>? type,
    required SunnyObservableList<T> list,
    this.primary,
    this.showDividers = false,
    this.diffEquality = const DiffEquality(),
    this.keyGenerator,
    this.onChange,
    this.useStateTracker = true,
    this.reverse = false,
    this.refreshWithPage,
    this.scrollController,
    this.padding,
    this.animationType = AnimationType.size,
    this.dismissAnimationType = AnimationType.slide,
    this.shrinkWrap,
    this.emptyState,
    this.primarySwipeActions = const [],
    this.secondarySwipeActions = const [],
    this.header,
  })  : state = type ?? (() => SunnyListState()),
        _start = [...list],
        vstream = (() => list.changeStream.replacements()),
        super(key: Key(id));

  @override
  State<SunnyList<T?>> createState() {
    return state();
  }
}

SunnyListWidgetBuilder<T> hideWhenEmpty<T>(Getter<Widget> widget) {
  return (context, list) {
    return list.isNotEmpty ? widget() : emptyBox;
  };
}

SunnyListWidgetBuilder<T> listHeader<T>(Widget widget,
    {bool hideWhenEmpty = true}) {
  return (context, list) {
    return (hideWhenEmpty == true && list.isNotEmpty) ? widget : emptyBox;
  };
}

SunnyListWidgetBuilder<T> sunnyHeader<T>(String headerText) {
  return (context, list) {
    return list.isNotEmpty ? SunnyListHeader(headerText) : emptyBox;
  };
}

abstract class SunnyListBaseState<T> extends BaseState<SunnyList<T>> {
  SunnyListBaseState();

  GlobalKey? get globalKey;

  ListTileBuilder<T> get itemBuilder =>
      widget.itemBuilder ?? nullPointer("No item builder provided!");

  late SingleValue<List<T>> _trackedValue;

  List<T>? get trackedList => _trackedValue.value;

  late KeyGenerator<T> _keyGenerator;

  /// A window of loaded list elements.  Certain list implementations will buffer the visible elements
  final _window = <int, SunnyListItemState<T>>{};

  late Map<Key, int> keyToIndex;

  late int size;

  int rebuilds = 0;

  bool get isNotEmpty => trackedList?.isNotEmpty == true;

  bool get isEmpty => trackedList?.isNotEmpty != true;

  bool get useStateTracker => widget.useStateTracker;

  @override
  String get loggerName => "${runtimeType.simpleName}.${widget.id}";

  DiffEquality get diffDelegator => widget.diffEquality;

  String? tryGetKey(int index, T item) {
    try {
      return _keyGenerator(item);
    } catch (e) {
      return "$index";
    }
  }

  void calculateKeys() {
    var i = 0;
    keyToIndex = Map.fromEntries(
      trackedList!.map(
        (item) => MapEntry(buildKey(i, item), i++),
      ),
    );
  }

  /// Builds a unique key for each list item that will be attached during the build process
  Key buildKey(int index, T item) {
    return PageStorageKey("list-item-${tryGetKey(index, item)}");
  }

  SyncStream<ListDiffs<T>>? _stream;

  StreamSubscription? _sub;

  /// Used because cancelling a future returns a Future and we want to drop any updates while we're disposing.
  bool disposing = false;

  @override
  @mustCallSuper
  void initState() {
    super.initState();
    _keyGenerator = widget.keyGenerator ?? _defaultKeyGen();
    _trackedValue = SingleValue<List<T>>(widget._start ?? <T>[]);
    _setupStream();
    size = _trackedValue.value!.length;
    if (widget.refreshWithPage != false) {
//      context.onPageRefresh(this.reset);
    }
  }

  @override
  void dispose() {
    log.finer("Disposing ${widget.id}");
    disposing = true;
    _sub?.cancel().then((_) {
      log.finer("-- Subscription cancelled");
      _stream?.dispose().then((_) {
        log.finer("-- Underlying stream closed");
      });
    }).ignore();

    super.dispose();
  }

  Future _setupStream() async {
    final Stream<Iterable<T>> _startStream = widget._start == null
        ? widget.vstream().flatten() as Stream<Iterable<T>>
        : widget.vstream().after as Stream<Iterable<T>>;

    /// asyncMap will ensure that previous operations are complete before starting the next.
    _stream = _startStream.map((iter) => iter.toList()).asyncMap((_) async {
      if (disposing) return ListDiffs<T>.empty();
      rebuilds++;
      final diff = await _calculateDiff(_);
      if (diff.isNotEmpty) {
        _onListChange(diff);
      }

      return diff;
    }).toSyncStream(ListDiffs.empty(), null, "listWidget[${widget.id}]");

    /// This stream listen does not listen to the first element in the list, only the subsequent ones
    _sub = _stream!.listen((_) {});
  }

  void insertItem(State? state, int? index, T itemToInsert);

  void removeItem(State? state, int? index, T itemToRemove);

  void updateItem(State state, int index, T oldItem, T newItem);

  void moveItem(State state, int fromIndex, int toIndex, T itemToMove);

  Widget onNotFound(
          BuildContext context, int index, Animation<double> animation) =>
      emptyBox;

  Widget buildItem(
      BuildContext context, int idx, T item, Animation<double> animation) {
    final innerItemBuilder = (BuildContext context) {
      final item = Lists.getOrNull(trackedList, idx);
      if (item == null) return emptyBox;
      final itemTile = itemBuilder(context, item, idx, animation);
      final itemTileWithSwipes = hasSwipeActions
          ? Slidable(
              closeOnScroll: true,
              actionExtentRatio: 0.20,
              showAllActionsThreshold: 0.3,
              child: itemTile,
              actions: [
                ...widget.primarySwipeActions
                    .map((swipeBuilder) => swipeBuilder(context, item)),
              ],
              secondaryActions: [
                ...widget.secondarySwipeActions
                    .map((swipeBuilder) => swipeBuilder(context, item)),
              ],
              actionPane: SlidableBehindActionPane(),
            )
          : itemTile;

      if (idx > 0 && widget.showDividers) {
        return Container(
          child: itemTileWithSwipes,
          decoration:
              BoxDecoration(border: Border(top: SunnyDeco.borderSideLight)),
        );
      } else {
        return itemTileWithSwipes;
      }
    };

    return useStateTracker
        ? SunnyListItem<T>(
            onInit: (state) {
              _window[idx] = state;
            },
            onDispose: (state) {
              _window.remove(idx);
            },
            key: buildKey(idx, item),
            list: trackedList,
            listId: widget.id,
            index: idx,
            itemBuilder: innerItemBuilder,
          )
        : innerItemBuilder(context);
  }

  Widget buildCompleteItem(
      BuildContext context, int idx, Animation<double> animation) {
    final item = Lists.getOrNull(trackedList, idx);
    if (item == null) return onNotFound(context, idx, animation);
    return animated(
      widget.animationType,
      name: "build-item-${widget.id}-$idx",
      animation: animation,
      child: buildItem(context, idx, item, animation),
    );
  }

  void _applyChange(ListDiffs<T> changes, ListDiff<T> listChange,
      [Iterable<SunnyListItemState<T>>? itemState]) {
    final state = globalKey?.currentState;
    if (globalKey != null && state == null) {
      log.info("ignoring updates because we have no state");
    }

    switch (listChange.type) {
      case ListDiffType.insert:
        log.infoJson('insert', {
          'index': listChange.index,
          'size': listChange.insert.size,
          'syncedIndexes': itemState?.length,
        });
        size += listChange.insert.size;
        for (final item in listChange.insert.items) {
          try {
            if (disposing) return;
            insertItem(state, listChange.insert.index, item);
          } catch (e) {
            log.warning("Problem inserting item: $e");
          }
        }
        break;
      case ListDiffType.delete:
        log.infoJson('delete', {
          'index': listChange.index,
          'size': listChange.delete.size,
          'syncedIndexes': itemState?.length,
        });
        size -= listChange.delete.size;
        for (final i in listChange.delete.indexes) {
          try {
            if (disposing) return;
            removeItem(
                state, listChange.delete.index, listChange.args.original[i]);
          } catch (e) {
            log.warning('Problem removing item: $e');
          }
        }

        break;
      case ListDiffType.replace:
        log.infoJson('replace', {
          'index': listChange.index,
          'size': listChange.change.size,
          'syncedIndexes': itemState?.length,
        });
        var indexes2 = [...listChange.change.indexes];
        var i = 0;
        for (final idx in indexes2) {
          final listIdx = idx;
          try {
            if (disposing) return;
            if (listIdx >= size) {
              log.info('Inserting item at $listIdx');
              insertItem(state, listIdx, listChange.change.items[i]);
            }
            if (widget.useStateTracker == false) {
              var replacement = listChange.change.items[i];
              final T? old = changes.args.original.tryGet(listIdx);
              if (widget.diffEquality.areIdentical(old, replacement) != true) {
                //remove and add
                log.fine('Removing and re-adding $listIdx');
                if (old != null) {
                  removeItem(state, listIdx, old);
                }
                insertItem(state, listIdx, replacement);
              }
            }

            final itemState = _window[idx];
            if (itemState != null) {
              log.fine('Found state at $listIdx - calling onChange');
              itemState.onChange();
            } else {
              log.fine('No state found at $listIdx, ignoring');
            }
          } catch (e, stack) {
            log.warning('Problem refreshing index  $idx', e, stack);
          } finally {
            i++;
          }
        }
        break;
    }
  }

  Future<ListDiffs<T>> _calculateDiff(List<T> newList) async {
    try {
      final start = [...trackedList!];
      final _max = max(newList.length, trackedList!.length);
      final _min = min(newList.length, trackedList!.length);
      if (_max > 50 && _min > 0) {
        log.info(
            'Async diff because lists are too long: ${newList.length} and ${trackedList!.length}');
        final diff = await start.differencesAsync(
          [...newList],
          debugName: widget.id,
          equality: widget.diffEquality,
        );
        return diff;
      } else {
        final diff = start.differences(
          [...newList],
          equality: widget.diffEquality,
        );
        return diff;
      }
    } catch (e, stack) {
      log.severe('Error running diff', e, stack);
      rethrow;
    }
  }

  void _onListChange(ListDiffs<T> changes) {
    _trackedValue.update([...changes.newList]);

    for (final change in changes) {
      _applyChange(changes, change);
    }

    widget.onChange?.call(changes.newList);
  }

  bool get hasSwipeActions =>
      widget.primarySwipeActions.isNotEmpty == true ||
      widget.secondarySwipeActions.isNotEmpty == true;

  bool get isDisplayed => trackedList!.isNotEmpty == true;

  String get debugName => 'widget[${widget.id}]';

  @override
  String toString({DiagnosticLevel minLevel = DiagnosticLevel.info}) {
    var toString = debugName;
    if (minLevel == DiagnosticLevel.debug) {
      var i = 0;
      trackedList!.forEach((item) => toString += '[${i++}] = $item\n');
    }
    return toString;
  }
}

typedef ListItemCallback<T> = void Function(SunnyListItemState<T> item);

/// Renders for each item in a [SunnyList] - this allows us to invoke refreshes when the individual list item
/// has changed (instead of removing, and then readding)
class SunnyListItem<T> extends StatefulWidget {
  final WidgetBuilder itemBuilder;
  final int index;
  final String listId;
  final List<T>? list;
  final ListItemCallback<T>? onInit;
  final ListItemCallback<T>? onDispose;

  const SunnyListItem({
    required Key key,
    required this.list,
    required this.listId,
    required this.itemBuilder,
    required this.index,
    this.onInit,
    this.onDispose,
  }) : super(key: key);

  @override
  State<StatefulWidget> createState() {
    return SunnyListItemState<T>();
  }
}

class SunnyListItemState<T> extends BaseState<SunnyListItem<T>> {
  @override
  void initState() {
    super.initState();
    widget.onInit?.call(this);
    log.fine("tracking ${widget.index}");
  }

  @override
  void dispose() {
    widget.onDispose?.call(this);
    log.fine("disposing list-item at ${widget.index}");
    super.dispose();
  }

  onChange() {
    if (mounted) {
      setState(() {});
    }
  }

  @override
  Widget build(BuildContext context) {
    return widget.itemBuilder(context);
  }

  @override
  String get loggerName {
    return runtimeType.simpleName + '.' + T.simpleName;
  }
}

KeyGenerator<T> _defaultKeyGen<T>() => (final T input) {
      if (input == null) return null;
      if (input is DiffDelegate) {
        return '${input.diffKey}';
      } else if (input is HasMverseMeta) {
        return input.mkeyOrNull?.mxid ??
            illegalState("Null mxid - a valid key is needed");
      } else if (input is String) {
        return input;
      } else if (input is num) {
        return '$input';
      } else {
        return illegalState("Can't generate key using default logic for $T");
      }
    };

extension DeleteDiffExt<E> on DeleteDiff<E> {
  Iterable<E> get oldItems {
    return this.indexes.map((i) => args.original[i]);
  }
}

enum SunnyListAction { reset }

/// Allows customizable control of the list beyond manipulating the source streams.
class SunnyListController<T> extends VNotifier<SunnyListAction> {
  final VNotifier<SunnyListAction> _actions = VNotifier();

  void reset() {
    _actions(SunnyListAction.reset);
  }
}
